# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Tuple, TypeVar, Union

from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset5 import Opset5
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset6(Opset5):
    def __new__(cls):
        return Opset.__new__(cls, "", 6)

    T_Abs = TypeVar(
        "T_Abs",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Abs(self, X: T_Abs) -> T_Abs:
        r"""[🌐 Abs(6)](https://onnx.ai/onnx/operators/onnx__Abs.html#abs-6 "Online Documentation")


        Absolute takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the absolute is, y = abs(x), is applied to
        the tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Abs", 6, "")
        op = Op(self, "Abs", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Add = TypeVar("T_Add", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def Add(
        self, A: T_Add, B: T_Add, *, axis: Optional[int] = None, broadcast: int = 0
    ) -> T_Add:
        r"""[🌐 Add(6)](https://onnx.ai/onnx/operators/onnx__Add.html#add-6 "Online Documentation")


        Performs element-wise binary addition (with limited broadcast support).

        If necessary the right-hand-side argument will be broadcasted to match the
        shape of left-hand-side argument. When broadcasting is specified, the second
        tensor can either be of element size 1 (including a scalar tensor and any
        tensor with rank equal to or smaller than the first tensor), or having its
        shape as a contiguous subset of the first tensor's shape. The starting of the
        mutually equal shape is specified by the argument "axis", and if it is not set,
        suffix matching is assumed. 1-dim expansion doesn't work yet.

        For example, the following tensor shapes are supported (with broadcast=1):

          shape(A) = (2, 3, 4, 5), shape(B) = (,), i.e. B is a scalar tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (1, 1), i.e. B is an 1-element tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (5,)
          shape(A) = (2, 3, 4, 5), shape(B) = (4, 5)
          shape(A) = (2, 3, 4, 5), shape(B) = (3, 4), with axis=1
          shape(A) = (2, 3, 4, 5), shape(B) = (2), with axis=0

        Attribute `broadcast=1` needs to be passed to enable broadcasting.


        Args:
            A: First operand, should share the type with the second operand.

            B: Second operand. With broadcasting can be of smaller size than A. If
                broadcasting is disabled it should be of the same size.

            axis: If set, defines the broadcast dimensions. See doc for details.

            broadcast: Pass 1 to enable broadcasting
        """

        schema = get_schema("Add", 6, "")
        op = Op(self, "Add", schema)
        return op(*self._prepare_inputs(schema, A, B), axis=axis, broadcast=broadcast)

    T_BatchNormalization = TypeVar("T_BatchNormalization", DOUBLE, FLOAT, FLOAT16)

    def BatchNormalization(
        self,
        X: T_BatchNormalization,
        scale: T_BatchNormalization,
        B: T_BatchNormalization,
        mean: T_BatchNormalization,
        var: T_BatchNormalization,
        *,
        epsilon: float = 9.999999747378752e-06,
        is_test: int = 0,
        momentum: float = 0.8999999761581421,
        spatial: int = 1,
    ) -> Tuple[
        T_BatchNormalization,
        T_BatchNormalization,
        T_BatchNormalization,
        T_BatchNormalization,
        T_BatchNormalization,
    ]:
        r"""[🌐 BatchNormalization(6)](https://onnx.ai/onnx/operators/onnx__BatchNormalization.html#batchnormalization-6 "Online Documentation")


        Carries out batch normalization as described in the paper
        https://arxiv.org/abs/1502.03167. Depending on the mode it is being run,
        there are multiple cases for the number of outputs, which we list below:

        Output case #1: Y, mean, var, saved_mean, saved_var (training mode)
        Output case #2: Y (test mode)


        Args:
            X: Input data tensor from the previous operator; dimensions for image case
                are (N x C x H x W), where N is the batch size, C is the number of
                channels, and H and W are the height and the width of the data. For non
                image case, the dimensions are in the form of (N x C x D1 x D2 ... Dn),
                where N is the batch size.

            scale: The scale as a 1-dimensional tensor of size C to be applied to the
                output.

            B: The bias as a 1-dimensional tensor of size C to be applied to the output.

            mean: The running mean (training) or the estimated mean (testing) as a
                1-dimensional tensor of size C.

            var: The running variance (training) or the estimated variance (testing) as
                a 1-dimensional tensor of size C.

            epsilon: The epsilon value to use to avoid division by zero, default is
                1e-5f.

            is_test: If set to nonzero, run spatial batch normalization in test mode,
                default is 0.

            momentum: Factor used in computing the running mean and variance.e.g.,
                running_mean = running_mean * momentum + mean * (1 - momentum), default
                is 0.9f.

            spatial: If true, compute the mean and variance across all spatial elements
                If false, compute the mean and variance across per feature.Default is 1.
        """

        schema = get_schema("BatchNormalization", 6, "")
        op = Op(self, "BatchNormalization", schema)
        return op(
            *self._prepare_inputs(schema, X, scale, B, mean, var),
            epsilon=epsilon,
            is_test=is_test,
            momentum=momentum,
            spatial=spatial,
        )

    T1_Cast = TypeVar(
        "T1_Cast",
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_Cast: TypeAlias = Union[
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def Cast(self, input: T1_Cast, *, to: int) -> T2_Cast:
        r"""[🌐 Cast(6)](https://onnx.ai/onnx/operators/onnx__Cast.html#cast-6 "Online Documentation")


        The operator casts the elements of a given input tensor to a data type
        specified by the 'to' argument and returns an output tensor of the same size in
        the converted type. The 'to' argument must be one of the data types specified
        in the 'DataType' enum field in the TensorProto message.
        NOTE: Casting to and from strings is not supported yet.


        Args:
            input: Input tensor to be cast.

            to: The data type to which the elements of the input tensor are cast.
                Strictly must be one of the types from DataType enum in TensorProto
        """

        schema = get_schema("Cast", 6, "")
        op = Op(self, "Cast", schema)
        return op(*self._prepare_inputs(schema, input), to=to)

    T_Ceil = TypeVar("T_Ceil", DOUBLE, FLOAT, FLOAT16)

    def Ceil(self, X: T_Ceil) -> T_Ceil:
        r"""[🌐 Ceil(6)](https://onnx.ai/onnx/operators/onnx__Ceil.html#ceil-6 "Online Documentation")


        Ceil takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the ceil is, y = ceil(x), is applied to
        the tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Ceil", 6, "")
        op = Op(self, "Ceil", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Clip = TypeVar("T_Clip", DOUBLE, FLOAT, FLOAT16)

    def Clip(
        self,
        input: T_Clip,
        *,
        max: float = 3.4028234663852886e38,
        min: float = -3.4028234663852886e38,
    ) -> T_Clip:
        r"""[🌐 Clip(6)](https://onnx.ai/onnx/operators/onnx__Clip.html#clip-6 "Online Documentation")


        Clip operator limits the given input within an interval. The interval is
        specified with arguments 'min' and 'max'. They default to
        numeric_limits::lowest() and numeric_limits::max() respectively.


        Args:
            input: Input tensor whose elements to be clipped

            max: Maximum value, above which element is replaced by max

            min: Minimum value, under which element is replaced by min
        """

        schema = get_schema("Clip", 6, "")
        op = Op(self, "Clip", schema)
        return op(*self._prepare_inputs(schema, input), max=max, min=min)

    T_Div = TypeVar("T_Div", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def Div(
        self, A: T_Div, B: T_Div, *, axis: Optional[int] = None, broadcast: int = 0
    ) -> T_Div:
        r"""[🌐 Div(6)](https://onnx.ai/onnx/operators/onnx__Div.html#div-6 "Online Documentation")


        Performs element-wise binary division (with limited broadcast support).

        If necessary the right-hand-side argument will be broadcasted to match the
        shape of left-hand-side argument. When broadcasting is specified, the second
        tensor can either be of element size 1 (including a scalar tensor and any
        tensor with rank equal to or smaller than the first tensor), or having its
        shape as a contiguous subset of the first tensor's shape. The starting of the
        mutually equal shape is specified by the argument "axis", and if it is not set,
        suffix matching is assumed. 1-dim expansion doesn't work yet.

        For example, the following tensor shapes are supported (with broadcast=1):

          shape(A) = (2, 3, 4, 5), shape(B) = (,), i.e. B is a scalar tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (1, 1), i.e. B is an 1-element tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (5,)
          shape(A) = (2, 3, 4, 5), shape(B) = (4, 5)
          shape(A) = (2, 3, 4, 5), shape(B) = (3, 4), with axis=1
          shape(A) = (2, 3, 4, 5), shape(B) = (2), with axis=0

        Attribute `broadcast=1` needs to be passed to enable broadcasting.


        Args:
            A: First operand, should share the type with the second operand.

            B: Second operand. With broadcasting can be of smaller size than A. If
                broadcasting is disabled it should be of the same size.

            axis: If set, defines the broadcast dimensions. See doc for details.

            broadcast: Pass 1 to enable broadcasting
        """

        schema = get_schema("Div", 6, "")
        op = Op(self, "Div", schema)
        return op(*self._prepare_inputs(schema, A, B), axis=axis, broadcast=broadcast)

    T_Dropout = TypeVar("T_Dropout", DOUBLE, FLOAT, FLOAT16)

    def Dropout(
        self, data: T_Dropout, *, is_test: int = 0, ratio: float = 0.5
    ) -> Tuple[T_Dropout, T_Dropout]:
        r"""[🌐 Dropout(6)](https://onnx.ai/onnx/operators/onnx__Dropout.html#dropout-6 "Online Documentation")


        Dropout takes one input data (Tensor<float>) and produces two Tensor outputs,
        output (Tensor<float>) and mask (Tensor<bool>). Depending on whether it is in
        test mode or not, the output Y will either be a random dropout, or a simple
        copy of the input. Note that our implementation of Dropout does scaling in
        the training phase, so during testing nothing needs to be done.


        Args:
            data: The input data as Tensor.

            is_test: (int, default 0) if nonzero, run dropout in test mode where the
                output is simply Y = X.

            ratio: (float, default 0.5) the ratio of random dropout
        """

        schema = get_schema("Dropout", 6, "")
        op = Op(self, "Dropout", schema)
        return op(*self._prepare_inputs(schema, data), is_test=is_test, ratio=ratio)

    T_Elu = TypeVar("T_Elu", DOUBLE, FLOAT, FLOAT16)

    def Elu(self, X: T_Elu, *, alpha: float = 1.0) -> T_Elu:
        r"""[🌐 Elu(6)](https://onnx.ai/onnx/operators/onnx__Elu.html#elu-6 "Online Documentation")


        Elu takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the function `f(x) = alpha * (exp(x) - 1.) for x <
        0`, `f(x) = x for x >= 0`., is applied to the tensor elementwise.



        Args:
            X: (differentiable) Input tensor

            alpha: Coefficient of ELU.
        """

        schema = get_schema("Elu", 6, "")
        op = Op(self, "Elu", schema)
        return op(*self._prepare_inputs(schema, X), alpha=alpha)

    T_Exp = TypeVar("T_Exp", DOUBLE, FLOAT, FLOAT16)

    def Exp(self, input: T_Exp) -> T_Exp:
        r"""[🌐 Exp(6)](https://onnx.ai/onnx/operators/onnx__Exp.html#exp-6 "Online Documentation")


        Calculates the exponential of the given input tensor, element-wise.


        Args:
            input: Input tensor
        """

        schema = get_schema("Exp", 6, "")
        op = Op(self, "Exp", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Floor = TypeVar("T_Floor", DOUBLE, FLOAT, FLOAT16)

    def Floor(self, X: T_Floor) -> T_Floor:
        r"""[🌐 Floor(6)](https://onnx.ai/onnx/operators/onnx__Floor.html#floor-6 "Online Documentation")


        Floor takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the floor is, y = floor(x), is applied to
        the tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Floor", 6, "")
        op = Op(self, "Floor", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Gemm = TypeVar("T_Gemm", DOUBLE, FLOAT, FLOAT16)

    def Gemm(
        self,
        A: T_Gemm,
        B: T_Gemm,
        C: T_Gemm,
        *,
        alpha: float = 1.0,
        beta: float = 1.0,
        broadcast: int = 0,
        transA: int = 0,
        transB: int = 0,
    ) -> T_Gemm:
        r"""[🌐 Gemm(6)](https://onnx.ai/onnx/operators/onnx__Gemm.html#gemm-6 "Online Documentation")

        General Matrix multiplication:
        https://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms#Level_3
        Compute Y = alpha * A * B + beta * C, where input tensor A has
        dimension (M X K), input tensor B has dimension (K X N), input tensor C and
        output tensor Y have dimension (M X N).
        If attribute broadcast is non-zero, input tensor C will be broadcasted to match
        the dimension requirement. A will be transposed before doing the computation
        if attribute transA is non-zero, same for B and transB.


        Args:
            A: Input tensor A

            B: Input tensor B

            C: Input tensor C

            alpha: Scalar multiplier for the product of input tensors A * B, the default
                value is 1.0.

            beta: Scalar multiplier for input tensor C, the default value is 1.0.

            broadcast: Whether C should be broadcasted

            transA: Whether A should be transposed

            transB: Whether B should be transposed
        """

        schema = get_schema("Gemm", 6, "")
        op = Op(self, "Gemm", schema)
        return op(
            *self._prepare_inputs(schema, A, B, C),
            alpha=alpha,
            beta=beta,
            broadcast=broadcast,
            transA=transA,
            transB=transB,
        )

    T_HardSigmoid = TypeVar("T_HardSigmoid", DOUBLE, FLOAT, FLOAT16)

    def HardSigmoid(
        self, X: T_HardSigmoid, *, alpha: float = 0.20000000298023224, beta: float = 0.5
    ) -> T_HardSigmoid:
        r"""[🌐 HardSigmoid(6)](https://onnx.ai/onnx/operators/onnx__HardSigmoid.html#hardsigmoid-6 "Online Documentation")


        HardSigmoid takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the HardSigmoid function, y = max(0, min(1, alpha * x + beta)),
        is applied to the tensor elementwise.


        Args:
            X: (differentiable) Input tensor

            alpha: Value of alpha.

            beta: Value of beta.
        """

        schema = get_schema("HardSigmoid", 6, "")
        op = Op(self, "HardSigmoid", schema)
        return op(*self._prepare_inputs(schema, X), alpha=alpha, beta=beta)

    T_InstanceNormalization = TypeVar("T_InstanceNormalization", DOUBLE, FLOAT, FLOAT16)

    def InstanceNormalization(
        self,
        input: T_InstanceNormalization,
        scale: T_InstanceNormalization,
        B: T_InstanceNormalization,
        *,
        epsilon: float = 9.999999747378752e-06,
    ) -> T_InstanceNormalization:
        r"""[🌐 InstanceNormalization(6)](https://onnx.ai/onnx/operators/onnx__InstanceNormalization.html#instancenormalization-6 "Online Documentation")


        Carries out instance normalization as described in the paper
        https://arxiv.org/abs/1607.08022.

        y = scale * (x - mean) / sqrt(variance + epsilon) + B,
        where mean and variance are computed per instance per channel.



        Args:
            input: (differentiable) Input data tensor from the previous operator;
                dimensions for image case are (N x C x H x W), where N is the batch
                size, C is the number of channels, and H and W are the height and the
                width of the data. For non image case, the dimensions are in the form of
                (N x C x D1 x D2 ... Dn), where N is the batch size.

            scale: (differentiable) The input 1-dimensional scale tensor of size C.

            B: (differentiable) The input 1-dimensional bias tensor of size C.

            epsilon: The epsilon value to use to avoid division by zero.
        """

        schema = get_schema("InstanceNormalization", 6, "")
        op = Op(self, "InstanceNormalization", schema)
        return op(*self._prepare_inputs(schema, input, scale, B), epsilon=epsilon)

    T_LeakyRelu = TypeVar("T_LeakyRelu", DOUBLE, FLOAT, FLOAT16)

    def LeakyRelu(self, X: T_LeakyRelu, *, alpha: float = 0.009999999776482582) -> T_LeakyRelu:
        r"""[🌐 LeakyRelu(6)](https://onnx.ai/onnx/operators/onnx__LeakyRelu.html#leakyrelu-6 "Online Documentation")


        LeakyRelu takes input data (Tensor<T>) and an argument alpha, and produces one
        output data (Tensor<T>) where the function `f(x) = alpha * x for x < 0`,
        `f(x) = x for x >= 0`, is applied to the data tensor elementwise.


        Args:
            X: (differentiable) Input tensor

            alpha: Coefficient of leakage.
        """

        schema = get_schema("LeakyRelu", 6, "")
        op = Op(self, "LeakyRelu", schema)
        return op(*self._prepare_inputs(schema, X), alpha=alpha)

    T_Log = TypeVar("T_Log", DOUBLE, FLOAT, FLOAT16)

    def Log(self, input: T_Log) -> T_Log:
        r"""[🌐 Log(6)](https://onnx.ai/onnx/operators/onnx__Log.html#log-6 "Online Documentation")


        Calculates the natural log of the given input tensor, element-wise.


        Args:
            input: Input tensor
        """

        schema = get_schema("Log", 6, "")
        op = Op(self, "Log", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Max = TypeVar("T_Max", DOUBLE, FLOAT, FLOAT16)

    def Max(self, *data_0: T_Max) -> T_Max:
        r"""[🌐 Max(6)](https://onnx.ai/onnx/operators/onnx__Max.html#max-6 "Online Documentation")


        Element-wise max of each of the input tensors. All inputs and outputs must
        have the same shape and data type.


        Args:
            data_0: (variadic) List of tensors for Max.
        """

        schema = get_schema("Max", 6, "")
        op = Op(self, "Max", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    T_Mean = TypeVar("T_Mean", DOUBLE, FLOAT, FLOAT16)

    def Mean(self, *data_0: T_Mean) -> T_Mean:
        r"""[🌐 Mean(6)](https://onnx.ai/onnx/operators/onnx__Mean.html#mean-6 "Online Documentation")


        Element-wise mean of each of the input tensors. All inputs and outputs must
        have the same shape and data type.


        Args:
            data_0: (variadic) List of tensors for Mean.
        """

        schema = get_schema("Mean", 6, "")
        op = Op(self, "Mean", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    T_Min = TypeVar("T_Min", DOUBLE, FLOAT, FLOAT16)

    def Min(self, *data_0: T_Min) -> T_Min:
        r"""[🌐 Min(6)](https://onnx.ai/onnx/operators/onnx__Min.html#min-6 "Online Documentation")


        Element-wise min of each of the input tensors. All inputs and outputs must
        have the same shape and data type.


        Args:
            data_0: (variadic) List of tensors for Min
        """

        schema = get_schema("Min", 6, "")
        op = Op(self, "Min", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    T_Mul = TypeVar("T_Mul", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def Mul(
        self, A: T_Mul, B: T_Mul, *, axis: Optional[int] = None, broadcast: int = 0
    ) -> T_Mul:
        r"""[🌐 Mul(6)](https://onnx.ai/onnx/operators/onnx__Mul.html#mul-6 "Online Documentation")


        Performs element-wise binary multiplication (with limited broadcast support).

        If necessary the right-hand-side argument will be broadcasted to match the
        shape of left-hand-side argument. When broadcasting is specified, the second
        tensor can either be of element size 1 (including a scalar tensor and any
        tensor with rank equal to or smaller than the first tensor), or having its
        shape as a contiguous subset of the first tensor's shape. The starting of the
        mutually equal shape is specified by the argument "axis", and if it is not set,
        suffix matching is assumed. 1-dim expansion doesn't work yet.

        For example, the following tensor shapes are supported (with broadcast=1):

          shape(A) = (2, 3, 4, 5), shape(B) = (,), i.e. B is a scalar tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (1, 1), i.e. B is an 1-element tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (5,)
          shape(A) = (2, 3, 4, 5), shape(B) = (4, 5)
          shape(A) = (2, 3, 4, 5), shape(B) = (3, 4), with axis=1
          shape(A) = (2, 3, 4, 5), shape(B) = (2), with axis=0

        Attribute `broadcast=1` needs to be passed to enable broadcasting.


        Args:
            A: First operand, should share the type with the second operand.

            B: Second operand. With broadcasting can be of smaller size than A. If
                broadcasting is disabled it should be of the same size.

            axis: If set, defines the broadcast dimensions. See doc for details.

            broadcast: Pass 1 to enable broadcasting
        """

        schema = get_schema("Mul", 6, "")
        op = Op(self, "Mul", schema)
        return op(*self._prepare_inputs(schema, A, B), axis=axis, broadcast=broadcast)

    T_Neg = TypeVar("T_Neg", DOUBLE, FLOAT, FLOAT16, INT16, INT32, INT64, INT8)

    def Neg(self, X: T_Neg) -> T_Neg:
        r"""[🌐 Neg(6)](https://onnx.ai/onnx/operators/onnx__Neg.html#neg-6 "Online Documentation")


        Neg takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where each element flipped sign, y = -x, is applied to
        the tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Neg", 6, "")
        op = Op(self, "Neg", schema)
        return op(*self._prepare_inputs(schema, X))

    T_PRelu = TypeVar("T_PRelu", DOUBLE, FLOAT, FLOAT16)

    def PRelu(self, X: T_PRelu, slope: T_PRelu) -> T_PRelu:
        r"""[🌐 PRelu(6)](https://onnx.ai/onnx/operators/onnx__PRelu.html#prelu-6 "Online Documentation")



        PRelu takes input data (Tensor<T>) and slope tensor as input, and produces one
        output data (Tensor<T>) where the function `f(x) = slope * x for x < 0`,
        `f(x) = x for x >= 0`., is applied to the data tensor elementwise.



        Args:
            X: Input tensor

            slope: Slope tensor. If `Slope` is of size 1, the value is sharedacross
                different channels
        """

        schema = get_schema("PRelu", 6, "")
        op = Op(self, "PRelu", schema)
        return op(*self._prepare_inputs(schema, X, slope))

    T_Reciprocal = TypeVar("T_Reciprocal", DOUBLE, FLOAT, FLOAT16)

    def Reciprocal(self, X: T_Reciprocal) -> T_Reciprocal:
        r"""[🌐 Reciprocal(6)](https://onnx.ai/onnx/operators/onnx__Reciprocal.html#reciprocal-6 "Online Documentation")


        Reciprocal takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the reciprocal is, y = 1/x, is applied to
        the tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Reciprocal", 6, "")
        op = Op(self, "Reciprocal", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Relu = TypeVar("T_Relu", DOUBLE, FLOAT, FLOAT16)

    def Relu(self, X: T_Relu) -> T_Relu:
        r"""[🌐 Relu(6)](https://onnx.ai/onnx/operators/onnx__Relu.html#relu-6 "Online Documentation")


        Relu takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the rectified linear function, y = max(0, x), is applied to
        the tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Relu", 6, "")
        op = Op(self, "Relu", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Selu = TypeVar("T_Selu", DOUBLE, FLOAT, FLOAT16)

    def Selu(
        self,
        X: T_Selu,
        *,
        alpha: float = 1.6732631921768188,
        gamma: float = 1.0507010221481323,
    ) -> T_Selu:
        r"""[🌐 Selu(6)](https://onnx.ai/onnx/operators/onnx__Selu.html#selu-6 "Online Documentation")


        Selu takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the scaled exponential linear unit function,
        `y = gamma * (alpha * e^x - alpha) for x <= 0`, `y = gamma * x for x > 0`,
        is applied to the tensor elementwise.


        Args:
            X: (differentiable) Input tensor

            alpha: Coefficient of SELU default to 1.67326319217681884765625 (i.e.,
                float32 approximation of 1.6732632423543772848170429916717).

            gamma: Coefficient of SELU default to 1.05070102214813232421875 (i.e.,
                float32 approximation of 1.0507009873554804934193349852946).
        """

        schema = get_schema("Selu", 6, "")
        op = Op(self, "Selu", schema)
        return op(*self._prepare_inputs(schema, X), alpha=alpha, gamma=gamma)

    T_Sigmoid = TypeVar("T_Sigmoid", DOUBLE, FLOAT, FLOAT16)

    def Sigmoid(self, X: T_Sigmoid) -> T_Sigmoid:
        r"""[🌐 Sigmoid(6)](https://onnx.ai/onnx/operators/onnx__Sigmoid.html#sigmoid-6 "Online Documentation")


        Sigmoid takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the sigmoid function, y = 1 / (1 + exp(-x)), is applied to the
        tensor elementwise.


        Args:
            X: Input tensor
        """

        schema = get_schema("Sigmoid", 6, "")
        op = Op(self, "Sigmoid", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Sqrt = TypeVar("T_Sqrt", DOUBLE, FLOAT, FLOAT16)

    def Sqrt(self, X: T_Sqrt) -> T_Sqrt:
        r"""[🌐 Sqrt(6)](https://onnx.ai/onnx/operators/onnx__Sqrt.html#sqrt-6 "Online Documentation")


        Square root takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the square root is, y = x^0.5, is applied to
        the tensor elementwise. If x is negative, then it will return NaN.


        Args:
            X: Input tensor
        """

        schema = get_schema("Sqrt", 6, "")
        op = Op(self, "Sqrt", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Sub = TypeVar("T_Sub", DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64)

    def Sub(
        self, A: T_Sub, B: T_Sub, *, axis: Optional[int] = None, broadcast: int = 0
    ) -> T_Sub:
        r"""[🌐 Sub(6)](https://onnx.ai/onnx/operators/onnx__Sub.html#sub-6 "Online Documentation")


        Performs element-wise binary subtraction (with limited broadcast support).

        If necessary the right-hand-side argument will be broadcasted to match the
        shape of left-hand-side argument. When broadcasting is specified, the second
        tensor can either be of element size 1 (including a scalar tensor and any
        tensor with rank equal to or smaller than the first tensor), or having its
        shape as a contiguous subset of the first tensor's shape. The starting of the
        mutually equal shape is specified by the argument "axis", and if it is not set,
        suffix matching is assumed. 1-dim expansion doesn't work yet.

        For example, the following tensor shapes are supported (with broadcast=1):

          shape(A) = (2, 3, 4, 5), shape(B) = (,), i.e. B is a scalar tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (1, 1), i.e. B is an 1-element tensor
          shape(A) = (2, 3, 4, 5), shape(B) = (5,)
          shape(A) = (2, 3, 4, 5), shape(B) = (4, 5)
          shape(A) = (2, 3, 4, 5), shape(B) = (3, 4), with axis=1
          shape(A) = (2, 3, 4, 5), shape(B) = (2), with axis=0

        Attribute `broadcast=1` needs to be passed to enable broadcasting.


        Args:
            A: First operand, should share the type with the second operand.

            B: Second operand. With broadcasting can be of smaller size than A. If
                broadcasting is disabled it should be of the same size.

            axis: If set, defines the broadcast dimensions. See doc for details.

            broadcast: Pass 1 to enable broadcasting
        """

        schema = get_schema("Sub", 6, "")
        op = Op(self, "Sub", schema)
        return op(*self._prepare_inputs(schema, A, B), axis=axis, broadcast=broadcast)

    T_Sum = TypeVar("T_Sum", DOUBLE, FLOAT, FLOAT16)

    def Sum(self, *data_0: T_Sum) -> T_Sum:
        r"""[🌐 Sum(6)](https://onnx.ai/onnx/operators/onnx__Sum.html#sum-6 "Online Documentation")


        Element-wise sum of each of the input tensors. All inputs and outputs must
        have the same shape and data type.


        Args:
            data_0: (variadic) List of tensors for Sum.
        """

        schema = get_schema("Sum", 6, "")
        op = Op(self, "Sum", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    T_Tanh = TypeVar("T_Tanh", DOUBLE, FLOAT, FLOAT16)

    def Tanh(self, input: T_Tanh) -> T_Tanh:
        r"""[🌐 Tanh(6)](https://onnx.ai/onnx/operators/onnx__Tanh.html#tanh-6 "Online Documentation")


        Calculates the hyperbolic tangent of the given input tensor element-wise.


        Args:
            input: Input tensor
        """

        schema = get_schema("Tanh", 6, "")
        op = Op(self, "Tanh", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Tile = TypeVar(
        "T_Tile",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_Tile: TypeAlias = INT64

    def Tile(self, input: T_Tile, repeats: T1_Tile) -> T_Tile:
        r"""[🌐 Tile(6)](https://onnx.ai/onnx/operators/onnx__Tile.html#tile-6 "Online Documentation")

        Constructs a tensor by tiling a given tensor.
        This is the same as function `tile` in Numpy, but no broadcast.
        For example A = [[1, 2], [3, 4]], B = [1, 2], tile(A, B) = [[1, 2, 1, 2], [3, 4, 3, 4]]


        Args:
            input: Input tensor of any shape.

            repeats: 1D int64 tensor of the same length as input's dimension number,
                includes numbers of repeated copies along input's dimensions.
        """

        schema = get_schema("Tile", 6, "")
        op = Op(self, "Tile", schema)
        return op(*self._prepare_inputs(schema, input, repeats))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence

import numpy as np
import onnx

from onnxscript import ir, tensor

if TYPE_CHECKING:
    from onnxscript._internal import converter

# Conversions from python values to ONNX are used by both the script converter as well
# as the eager-mode runtime and both need to be consistent. The script converter converts
# python values into ONNX TensorProto, while the runtime converts python values into
# ONNXScript runtime's value-representation (based on Tensor).

_REPEATED_ATTRIBUTE_TYPES = frozenset(
    {
        ir.AttributeType.FLOATS,
        ir.AttributeType.INTS,
        ir.AttributeType.STRINGS,
        ir.AttributeType.TENSORS,
        ir.AttributeType.GRAPHS,
        ir.AttributeType.SPARSE_TENSORS,
        ir.AttributeType.TYPE_PROTOS,
    }
)


def pyvalue_to_onnx_attribute(
    key: str,
    value: Any,
    name_generator: Callable[[], str],
    attr_type: ir.AttributeType | None = None,
) -> ir.Attr:
    """Helper function to create an ONNX AttributeProto.

    * Empty lists can be attribute values, provided the attribute type is specified
    and is a list type.
    * Scalar-values like 1.0 as well as lists like [1, -1] to be specified
    when the attribute type is TensorProto by automatically converting the value
    into a 0-D or 1-D tensor respectively.
    """
    # TODO(justinchuby): Remove this function and use onnx-ir directly.
    if isinstance(value, list) and not value:
        # Empty list value:
        if attr_type is None:
            raise ValueError("Attribute type must be specified for empty list value.")
        if attr_type not in _REPEATED_ATTRIBUTE_TYPES:
            raise ValueError("Empty list value is only allowed for repeated attribute types.")
        return ir.Attr(name=key, type=attr_type, value=[])
    elif attr_type == ir.AttributeType.TENSOR and not isinstance(value, onnx.TensorProto):
        return ir.AttrTensor(name=key, value=ir.tensor(value, name=name_generator()))
    else:
        return ir.convenience.convert_attribute(key, value, attr_type=attr_type)


# Utilities to convert python values into onnxscript tensors.


def _promotable(x) -> bool:
    """Checks if a runtime parameter value needs to be promoted into an onnxscript value.
    This is the runtime-equivalent of the promotion of literal constants into ONNX values
    in the static converter.
    """
    if isinstance(x, (bool, int, float)):
        return True
    if isinstance(x, list) and x:
        # Note: This is meant to handle valid scenarios correctly. No attempt is
        # made yet to capture all invalid usages in runtime mode.
        return _promotable(x[0])
    return False


def _get_dtype(pyvalue):
    """Return np.dtype to use when converting a python value to an onnxscript tensor.
    Note that int constants are treated as int64, as that is the common type in ONNX
    for shape/index values.
    """
    if isinstance(pyvalue, bool):
        return np.bool_
    elif isinstance(pyvalue, int):
        return np.int64
    elif isinstance(pyvalue, float):
        return np.float32
    elif isinstance(pyvalue, list):
        if pyvalue:
            # TODO: What to do about lists with mixed value types, like [1, 2.0]?
            # Should at least produce an error/warning message.
            return _get_dtype(pyvalue[0])
        raise ValueError("Cannot determine target type for empty list")
    raise TypeError(f"Value of unexpected type {type(pyvalue)}")


def cast_pyvalue_to_os_tensor(pyvalue, dtype=None):
    """Promotes python values into onnxscript tensors.
    The optional argument dtype specifies the desired np.dtype of the tensor,
    used only when a non-standard onnxscript-value is promoted into one.
    """
    if _promotable(pyvalue):
        if dtype is None:
            dtype = _get_dtype(pyvalue)
        return tensor.Tensor(np.array(pyvalue, dtype=dtype))
    return pyvalue


def cast_inputs(
    get_type_info: Callable[[Any], Any],
    cast: Callable[[Any, Any], Any],
    op_signature: ir.schemas.OpSignature | None,
    args,
) -> tuple[Any, ...]:
    """Uses schema specification to support a limited form of auto-casting.

    * Scalars are promoted to tensors.
    * Further. they are cast to the required type when used in ops with other
    tensor inputs that are required to be of same type.
    Thus, in "A+1" or "Add(A, 1)", the value 1 will be converted to the same
    type as A.

    This is used by the converter in a static-mode, as well as by the eager-mode
    execution in a dynamic-mode.
    """
    if op_signature is None:
        # Either an error or a custom op.
        # No checks/casts in this case.
        return tuple(cast(x, None) for x in args)

    # Filter to get only input parameters (not AttributeParameters)
    expected_inputs = op_signature.inputs
    # We make two passes. In the first pass, we identify known type-bindings for
    # type-variables: eg., {'T1' : np.float32, 'T2' : np.int32}.
    # In the second pass, we use these bindings to cast scalar-values to
    # tensors of appropriate types. The two passes are needed to handle cases
    # like "Add(1, X)" where 1 must be cast to the same type as X.
    type_bindings: dict[Optional[str], np.dtype] = {}
    args_typevars: list[tuple[str, Optional[str]]] = []
    for i, x in enumerate(args):
        if i < len(expected_inputs):
            expected = expected_inputs[i]
        elif expected_inputs[-1].variadic:
            expected = expected_inputs[-1]
            if not expected.homogeneous:
                args_typevars.append((x, None))
                continue
        else:
            raise ValueError(
                f"Number of actual parameters {len(args)} "
                f"exceeds number of formal parameters {len(expected_inputs)}."
            )
        typevar = expected.type_constraint.name
        if "(" not in typevar:
            # typevar is an identifier, like "T"
            typeinfo = get_type_info(x)
            if typeinfo is not None:
                type_bindings[typevar] = typeinfo
        args_typevars.append((x, typevar))
    cast_args = [cast(x, type_bindings.get(typevar)) for x, typevar in args_typevars]
    return tuple(cast_args)


def dynamic_cast_inputs(op_signature: ir.schemas.OpSignature, args):
    """Used for autocast during eager-mode execution."""

    def get_type_info(x):
        return x.dtype if isinstance(x, tensor.Tensor) else None

    return cast_inputs(get_type_info, cast_pyvalue_to_os_tensor, op_signature, args)


def static_cast_inputs(
    converter_: converter.Converter,
    op_signature: Optional[ir.schemas.OpSignature],
    args: Sequence[Optional[ir.Value]],
) -> tuple[str, ...]:
    """Used for autocast during script-translation.
    This is meant to transform expressions like "Add(X, 1)" to "Add(X, CastLike(1, X))"
    Polymorphic constants (like 0 and 1) are cast to the type of other operands as needed.
    """

    def get_type_info(x: Optional[ir.Value]) -> Optional[ir.Value]:
        """Returns x back if x can serve as the target-type for a cast (as the second
        argument of CastLike) and None otherwise. In the expression "Add(X, 1), 1 is
        castable, while X can serve as the target-type.
        """
        return None if x is None or converter_._is_castable(x.name) else x  # pylint: disable=protected-access

    def cast_like(x: Optional[ir.Value], y: Optional[ir.Value]) -> Optional[str]:
        if x is None:
            return None
        if converter_._is_castable(x.name) and y is not None:  # pylint: disable=protected-access
            # Polymorphic constant x is cast to the type of y:
            x_cast = converter_._generate_unique_name(f"{x.name}_cast")  # pylint: disable=protected-access
            return converter_._emit1([x_cast], "CastLike", [x, y])  # pylint: disable=protected-access
        return x

    return cast_inputs(get_type_info, cast_like, op_signature, args)



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Tuple, TypeVar, Union

from onnx import TensorProto
from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset19 import Opset19
from onnxscript.onnx_types import (
    BFLOAT16,
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT8E4M3FN,
    FLOAT8E4M3FNUZ,
    FLOAT8E5M2,
    FLOAT8E5M2FNUZ,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset20(Opset19):
    def __new__(cls):
        return Opset.__new__(cls, "", 20)

    T1_AffineGrid = TypeVar("T1_AffineGrid", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    T2_AffineGrid: TypeAlias = INT64

    def AffineGrid(
        self, theta: T1_AffineGrid, size: T2_AffineGrid, *, align_corners: int = 0
    ) -> T1_AffineGrid:
        r"""[🌐 AffineGrid(20)](https://onnx.ai/onnx/operators/onnx__AffineGrid.html#affinegrid-20 "Online Documentation")


        Generates a 2D or 3D flow field (sampling grid), given a batch of affine matrices theta
        (https://pytorch.org/docs/stable/generated/torch.nn.functional.affine_grid.html).
        An affine matrix `theta` is applied to a position tensor represented in its homogeneous expression. Here is an example in 3D:
        ::

            [r00, r01, r02, t0]   [x]   [x']
            [r10, r11, r12, t1] * [y] = [y']
            [r20, r21, r22, t2]   [z]   [z']
            [0,   0,   0,   1 ]   [1]   [1 ]


        where `(x, y, z)` is the position in the original space, `(x', y', z')` is the position in the output space.
        The last row is always `[0, 0, 0, 1]` and is not stored in the affine matrix. Therefore we have `theta` of shape `(N, 2, 3)` for 2D or `(N, 3, 4)` for 3D.

        Input `size` is used to define grid of positions evenly spaced in the original 2D or 3D space, with dimensions ranging from `-1` to `1`.
        The output `grid` contains positions in the output space.

        When `align_corners=1`, consider `-1` and `1` to refer to the centers of the corner pixels (mark `v` in illustration).
        ::

            v            v            v            v
            |-------------------|------------------|
            -1                  0                  1


        When `align_corners=0`, consider `-1` and `1` to refer to the outer edge of the corner pixels.
        ::

                v        v         v         v
            |------------------|-------------------|
            -1                 0                   1




        Args:
            theta: (non-differentiable) input batch of affine matrices with shape (N, 2,
                3) for 2D or (N, 3, 4) for 3D

            size: (non-differentiable) the target output image size (N, C, H, W) for 2D
                or (N, C, D, H, W) for 3D

            align_corners: if align_corners=1, consider -1 and 1 to refer to the centers
                of the corner pixels. if align_corners=0, consider -1 and 1 to refer to
                the outer edge the corner pixels.
        """

        schema = get_schema("AffineGrid", 20, "")
        op = Op(self, "AffineGrid", schema)
        return op(*self._prepare_inputs(schema, theta, size), align_corners=align_corners)

    T1_ConstantOfShape: TypeAlias = INT64

    T2_ConstantOfShape: TypeAlias = Union[
        BFLOAT16,
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        FLOAT8E4M3FN,
        FLOAT8E4M3FNUZ,
        FLOAT8E5M2,
        FLOAT8E5M2FNUZ,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def ConstantOfShape(
        self, input: T1_ConstantOfShape, *, value: Optional[TensorProto] = None
    ) -> T2_ConstantOfShape:
        r"""[🌐 ConstantOfShape(20)](https://onnx.ai/onnx/operators/onnx__ConstantOfShape.html#constantofshape-20 "Online Documentation")


        Generate a tensor with given value and shape.


        Args:
            input: 1D tensor. The shape of the expected output tensor. If empty tensor
                is given, the output would be a scalar. All values must be >= 0.

            value: (Optional) The value of the output elements.Should be a one-element
                tensor. If not specified, it defaults to a tensor of value 0 and
                datatype float32
        """

        schema = get_schema("ConstantOfShape", 20, "")
        op = Op(self, "ConstantOfShape", schema)
        return op(*self._prepare_inputs(schema, input), value=value)

    T1_DFT = TypeVar("T1_DFT", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    T2_DFT = TypeVar("T2_DFT", INT32, INT64)

    def DFT(
        self,
        input: T1_DFT,
        dft_length: Optional[T2_DFT] = None,
        axis: Optional[INT64] = None,
        *,
        inverse: int = 0,
        onesided: int = 0,
    ) -> T1_DFT:
        r"""[🌐 DFT(20)](https://onnx.ai/onnx/operators/onnx__DFT.html#dft-20 "Online Documentation")

        Computes the discrete Fourier Transform (DFT) of the input.

        Assuming the input has shape `[M, N]`, where `N` is the dimension over which the
        DFT is computed and `M` denotes the conceptual "all other dimensions,"
        the DFT `y[m, k]` of shape `[M, N]` is defined as

        $$y[m, k] = \sum_{n=0}^{N-1} e^{-2 \pi j \frac{k n}{N} } x[m, n] ,$$

        and the inverse transform is defined as

        $$x[m, n] = \frac{1}{N} \sum_{k=0}^{N-1} e^{2 \pi j \frac{k n}{N} } y[m, k] ,$$

        where $j$ is the imaginary unit.

        The actual shape of the output is specified in the "output" section.

        Reference: https://docs.scipy.org/doc/scipy/tutorial/fft.html


        Args:
            input: (non-differentiable) For real input, the following shape is expected:
                `[signal_dim0][signal_dim1][signal_dim2]...[signal_dimN][1]`. For
                complex input, the following shape is expected:
                `[signal_dim0][signal_dim1][signal_dim2]...[signal_dimN][2]`. The final
                dimension represents the real and imaginary parts of the value in that
                order.

            dft_length: (optional, non-differentiable) The length of the signal as a
                scalar. If greater than the axis dimension, the signal will be
                zero-padded up to `dft_length`. If less than the axis dimension, only
                the first `dft_length` values will be used as the signal.

            axis: (optional, non-differentiable) The axis as a scalar on which to
                perform the DFT. Default is `-2` (last signal axis). Negative value
                means counting dimensions from the back. Accepted range is $[-r, -2]
                \cup [0, r-2]$ where `r = rank(input)`. The last dimension is for
                representing complex numbers and thus is an invalid axis.

            inverse: Whether to perform the inverse discrete Fourier Transform. Default
                is 0, which corresponds to `false`.

            onesided: If `onesided` is `1` and input is real, only values for `k` in
                `[0, 1, 2, ..., floor(n_fft/2) + 1]` are returned because the
                real-to-complex Fourier transform satisfies the conjugate symmetry,
                i.e., `X[m, k] = X[m, n_fft-k]*`, where `m` denotes "all other
                dimensions" DFT was not applied on. If the input tensor is complex,
                onesided output is not possible. Value can be `0` or `1`. Default is
                `0`.
        """

        schema = get_schema("DFT", 20, "")
        op = Op(self, "DFT", schema)
        return op(
            *self._prepare_inputs(schema, input, dft_length, axis),
            inverse=inverse,
            onesided=onesided,
        )

    T_Gelu = TypeVar("T_Gelu", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    def Gelu(self, X: T_Gelu, *, approximate: str = "none") -> T_Gelu:
        r"""[🌐 Gelu(20)](https://onnx.ai/onnx/operators/onnx__Gelu.html#gelu-20 "Online Documentation")


        Gelu takes one input data (Tensor<T>) and produces one
        output data (Tensor<T>) where the gaussian error linear units function,
        $y = 0.5 * x * (1 + erf(x/sqrt(2)))$ is applied to the tensor elementwise.
        If the attribute "approximate" is set to "tanh", the function estimation,
        $y = 0.5 * x * (1 + Tanh(sqrt(2/\pi) * (x + 0.044715 * x^3)))$ is used and applied
        to the tensor elementwise.



        Args:
            X: (differentiable) Input tensor

            approximate: Gelu approximation algorithm: `"tanh"`,
                `"none"`(default).`"none"`: do not use approximation.`"tanh"`: use tanh
                approximation.
        """

        schema = get_schema("Gelu", 20, "")
        op = Op(self, "Gelu", schema)
        return op(*self._prepare_inputs(schema, X), approximate=approximate)

    T1_GridSample = TypeVar(
        "T1_GridSample",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_GridSample = TypeVar("T2_GridSample", DOUBLE, FLOAT, FLOAT16)

    def GridSample(
        self,
        X: T1_GridSample,
        grid: T2_GridSample,
        *,
        align_corners: int = 0,
        mode: str = "linear",
        padding_mode: str = "zeros",
    ) -> T1_GridSample:
        r"""[🌐 GridSample(20)](https://onnx.ai/onnx/operators/onnx__GridSample.html#gridsample-20 "Online Documentation")


        Given an input `X` and a flow-field `grid`, computes the output `Y` using `X` values and pixel locations from the `grid`.
        For spatial input `X` with shape (N, C, H, W), the `grid` will have shape (N, H_out, W_out, 2),
        the output `Y` will have shape (N, C, H_out, W_out). For volumetric input `X` with shape (N, C, D, H, W),
        the `grid` will have shape (N, D_out, H_out, W_out, 3), the output `Y` will have shape (N, C, D_out, H_out, W_out).
        More generally, for an input `X` of rank r+2 with shape (N, C, d1, d2, ..., dr),
        the `grid` will have shape (N, D1_out, D2_out, ..., Dr_out, r), the output `Y` will have shape (N, C, D1_out, D2_out, ..., Dr_out).

        The tensor `X` contains values at centers of square pixels (voxels, etc) locations such as (n, c, d1_in, d2_in, ..., dr_in).
        The (n, d1_out, d2_out, ..., dr_out, :) values from the tensor `grid` are the normalized positions for interpolating the values
        at the (n, c, d1_out, d2_out, ..., dr_out) locations from the output tensor `Y` using a specified interpolation method (the mode)
        and a padding mode (for `grid` positions falling outside the 2-dimensional image).

        For example, the values in `grid[n, h_out, w_out, :]` are size-2 vectors specifying normalized positions in the 2-dimensional space of `X`.
        They are used to interpolate output values of `Y[n, c, h_out, w_out]`.

        The GridSample operator is often used in doing grid generator and sampler in the
        [Spatial Transformer Networks](https://arxiv.org/abs/1506.02025).
        See also in [torch.nn.functional.grid_sample](https://pytorch.org/docs/stable/generated/torch.nn.functional.grid_sample.html).


        Args:
            X: (differentiable) Input tensor of rank r+2 that has shape (N, C, D1, D2,
                ..., Dr), where N is the batch size, C is the number of channels, D1,
                D2, ..., Dr are the spatial dimensions.

            grid: (non-differentiable) Input offset of shape (N, D1_out, D2_out, ...,
                Dr_out, r), where D1_out, D2_out, ..., Dr_out are the spatial dimensions
                of the grid and output, and r is the number of spatial dimensions. Grid
                specifies the sampling locations normalized by the input spatial
                dimensions. Therefore, it should have most values in the range of [-1,
                1]. If the grid has values outside the range of [-1, 1], the
                corresponding outputs will be handled as defined by padding_mode.
                Following computer vision convention, the coordinates in the length-r
                location vector are listed from the innermost tensor dimension to the
                outermost, the opposite of regular tensor indexing.

            align_corners: If align_corners=1, the extrema (-1 and 1) are considered as
                referring to the center points of the input's corner pixels (voxels,
                etc.). If align_corners=0, they are instead considered as referring to
                the corner points of the input's corner pixels (voxels, etc.), making
                the sampling more resolution agnostic.

            mode: Three interpolation modes: linear (default), nearest and cubic. The
                "linear" mode includes linear and N-linear interpolation modes depending
                on the number of spatial dimensions of the input tensor (i.e. linear for
                1 spatial dimension, bilinear for 2 spatial dimensions, etc.). The
                "cubic" mode also includes N-cubic interpolation modes following the
                same rules. The "nearest" mode rounds to the nearest even index when the
                sampling point falls halfway between two indices.

            padding_mode: Support padding modes for outside grid values:
                `zeros`(default), `border`, `reflection`. zeros: use 0 for out-of-bound
                grid locations, border: use border values for out-of-bound grid
                locations, reflection: use values at locations reflected by the border
                for out-of-bound grid locations. If index 0 represents the margin pixel,
                the reflected value at index -1 will be the same as the value at index
                1. For location far away from the border, it will keep being reflected
                until becoming in bound. If pixel location x = -3.5 reflects by border
                -1 and becomes x' = 1.5, then reflects by border 1 and becomes x'' =
                0.5.
        """

        schema = get_schema("GridSample", 20, "")
        op = Op(self, "GridSample", schema)
        return op(
            *self._prepare_inputs(schema, X, grid),
            align_corners=align_corners,
            mode=mode,
            padding_mode=padding_mode,
        )

    T1_ImageDecoder: TypeAlias = UINT8

    T2_ImageDecoder: TypeAlias = UINT8

    def ImageDecoder(
        self, encoded_stream: T1_ImageDecoder, *, pixel_format: str = "RGB"
    ) -> T2_ImageDecoder:
        r"""[🌐 ImageDecoder(20)](https://onnx.ai/onnx/operators/onnx__ImageDecoder.html#imagedecoder-20 "Online Documentation")

        Loads and decodes and image from a file. If it can't decode for any reason (e.g. corrupted encoded
        stream, invalid format, it will return an empty matrix).
        The following image formats are supported:
        * BMP
        * JPEG (note: Lossless JPEG support is optional)
        * JPEG2000
        * TIFF
        * PNG
        * WebP
        * Portable image format (PBM, PGM, PPM, PXM, PNM)
        Decoded images follow a channel-last layout: (Height, Width, Channels).
        **JPEG chroma upsampling method:**
        When upsampling the chroma components by a factor of 2, the pixels are linearly interpolated so that the
        centers of the output pixels are 1/4 and 3/4 of the way between input pixel centers.
        When rounding, 0.5 is rounded down and up at alternative pixels locations to prevent bias towards
        larger values (ordered dither pattern).
        Considering adjacent input pixels A, B, and C, B is upsampled to pixels B0 and B1 so that
        ::

            B0 = round_half_down((1/4) * A + (3/4) * B)
            B1 = round_half_up((3/4) * B + (1/4) * C)


        This method,  is the default chroma upsampling method in the well-established libjpeg-turbo library,
        also referred as "smooth" or "fancy" upsampling.


        Args:
            encoded_stream: (non-differentiable) Encoded stream

            pixel_format: Pixel format. Can be one of "RGB", "BGR", or "Grayscale".
        """

        schema = get_schema("ImageDecoder", 20, "")
        op = Op(self, "ImageDecoder", schema)
        return op(*self._prepare_inputs(schema, encoded_stream), pixel_format=pixel_format)

    T1_IsInf = TypeVar(
        "T1_IsInf",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        FLOAT8E4M3FN,
        FLOAT8E4M3FNUZ,
        FLOAT8E5M2,
        FLOAT8E5M2FNUZ,
    )

    T2_IsInf: TypeAlias = BOOL

    def IsInf(
        self, X: T1_IsInf, *, detect_negative: int = 1, detect_positive: int = 1
    ) -> T2_IsInf:
        r"""[🌐 IsInf(20)](https://onnx.ai/onnx/operators/onnx__IsInf.html#isinf-20 "Online Documentation")

        Map infinity to true and other values to false.

        Args:
            X: (non-differentiable) input

            detect_negative: (Optional) Whether map negative infinity to true. Default
                to 1 so that negative infinity induces true. Set this attribute to 0 if
                negative infinity should be mapped to false.

            detect_positive: (Optional) Whether map positive infinity to true. Default
                to 1 so that positive infinity induces true. Set this attribute to 0 if
                positive infinity should be mapped to false.
        """

        schema = get_schema("IsInf", 20, "")
        op = Op(self, "IsInf", schema)
        return op(
            *self._prepare_inputs(schema, X),
            detect_negative=detect_negative,
            detect_positive=detect_positive,
        )

    T1_IsNaN = TypeVar(
        "T1_IsNaN",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        FLOAT8E4M3FN,
        FLOAT8E4M3FNUZ,
        FLOAT8E5M2,
        FLOAT8E5M2FNUZ,
    )

    T2_IsNaN: TypeAlias = BOOL

    def IsNaN(self, X: T1_IsNaN) -> T2_IsNaN:
        r"""[🌐 IsNaN(20)](https://onnx.ai/onnx/operators/onnx__IsNaN.html#isnan-20 "Online Documentation")

        Returns which elements of the input are NaN.

        Args:
            X: (non-differentiable) input
        """

        schema = get_schema("IsNaN", 20, "")
        op = Op(self, "IsNaN", schema)
        return op(*self._prepare_inputs(schema, X))

    T_ReduceMax = TypeVar(
        "T_ReduceMax",
        BFLOAT16,
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT32,
        INT64,
        INT8,
        UINT32,
        UINT64,
        UINT8,
    )

    def ReduceMax(
        self,
        data: T_ReduceMax,
        axes: Optional[INT64] = None,
        *,
        keepdims: int = 1,
        noop_with_empty_axes: int = 0,
    ) -> T_ReduceMax:
        r"""[🌐 ReduceMax(20)](https://onnx.ai/onnx/operators/onnx__ReduceMax.html#reducemax-20 "Online Documentation")


        Computes the max of the input tensor's elements along the provided axes. The resulting
        tensor has the same rank as the input if `keepdims` equals 1. If `keepdims` equals 0, then
        the resulting tensor has the reduced dimension pruned. Input tensors of rank zero are
        valid. Reduction over an empty set of values yields minus infinity (if supported by the datatype) or the minimum value of the data type otherwise.


        If the input data type is Boolean, the comparison should consider `False < True`.

        The above behavior is similar to numpy, with the exception that numpy defaults `keepdims`
        to `False` instead of `True`.

        Args:
            data: (differentiable) An input tensor.

            axes: (optional, non-differentiable) Optional input list of integers, along
                which to reduce. The default is to reduce over empty axes. When axes is
                empty (either not provided or explicitly empty), behavior depends on
                'noop_with_empty_axes': reduction over all axes if
                'noop_with_empty_axes' is false, or no reduction is applied if
                'noop_with_empty_axes' is true (but other operations will be performed).
                Accepted range is [-r, r-1] where r = rank(data).

            keepdims: Keep the reduced dimension or not, default 1 means keep reduced
                dimension.

            noop_with_empty_axes: Defines behavior when axes is not provided or is
                empty. If false (default), reduction happens over all axes. If true, no
                reduction is applied, but other operations will be performed. For
                example, ReduceSumSquare acts as a vanilla Square.
        """

        schema = get_schema("ReduceMax", 20, "")
        op = Op(self, "ReduceMax", schema)
        return op(
            *self._prepare_inputs(schema, data, axes),
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )

    T_ReduceMin = TypeVar(
        "T_ReduceMin",
        BFLOAT16,
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT32,
        INT64,
        INT8,
        UINT32,
        UINT64,
        UINT8,
    )

    def ReduceMin(
        self,
        data: T_ReduceMin,
        axes: Optional[INT64] = None,
        *,
        keepdims: int = 1,
        noop_with_empty_axes: int = 0,
    ) -> T_ReduceMin:
        r"""[🌐 ReduceMin(20)](https://onnx.ai/onnx/operators/onnx__ReduceMin.html#reducemin-20 "Online Documentation")


        Computes the min of the input tensor's elements along the provided axes. The resulting
        tensor has the same rank as the input if `keepdims` equals 1. If `keepdims` equals 0, then
        the resulting tensor has the reduced dimension pruned. Input tensors of rank zero are
        valid. Reduction over an empty set of values yields plus infinity (if supported by the datatype) or the maximum value of the data type otherwise.


        If the input data type is Boolean, the comparison should consider `False < True`.

        The above behavior is similar to numpy, with the exception that numpy defaults `keepdims`
        to `False` instead of `True`.

        Args:
            data: (differentiable) An input tensor.

            axes: (optional, non-differentiable) Optional input list of integers, along
                which to reduce. The default is to reduce over empty axes. When axes is
                empty (either not provided or explicitly empty), behavior depends on
                'noop_with_empty_axes': reduction over all axes if
                'noop_with_empty_axes' is false, or no reduction is applied if
                'noop_with_empty_axes' is true (but other operations will be performed).
                Accepted range is [-r, r-1] where r = rank(data).

            keepdims: Keep the reduced dimension or not, default 1 means keep reduced
                dimension.

            noop_with_empty_axes: Defines behavior when axes is not provided or is
                empty. If false (default), reduction happens over all axes. If true, no
                reduction is applied, but other operations will be performed. For
                example, ReduceSumSquare acts as a vanilla Square.
        """

        schema = get_schema("ReduceMin", 20, "")
        op = Op(self, "ReduceMin", schema)
        return op(
            *self._prepare_inputs(schema, data, axes),
            keepdims=keepdims,
            noop_with_empty_axes=noop_with_empty_axes,
        )

    T1_RegexFullMatch: TypeAlias = STRING

    T2_RegexFullMatch: TypeAlias = BOOL

    def RegexFullMatch(
        self, X: T1_RegexFullMatch, *, pattern: Optional[str] = None
    ) -> T2_RegexFullMatch:
        r"""[🌐 RegexFullMatch(20)](https://onnx.ai/onnx/operators/onnx__RegexFullMatch.html#regexfullmatch-20 "Online Documentation")

        RegexFullMatch performs a full regex match on each element of the input tensor. If an element fully matches the regex pattern specified as an attribute, the corresponding element in the output is True and it is False otherwise. [RE2](https://github.com/google/re2/wiki/Syntax) regex syntax is used.

        Args:
            X: (non-differentiable) Tensor with strings to match on.

            pattern: Regex pattern to match on. This must be valid RE2 syntax.
        """

        schema = get_schema("RegexFullMatch", 20, "")
        op = Op(self, "RegexFullMatch", schema)
        return op(*self._prepare_inputs(schema, X), pattern=pattern)

    T_StringConcat: TypeAlias = STRING

    def StringConcat(self, X: T_StringConcat, Y: T_StringConcat) -> T_StringConcat:
        r"""[🌐 StringConcat(20)](https://onnx.ai/onnx/operators/onnx__StringConcat.html#stringconcat-20 "Online Documentation")

        StringConcat concatenates string tensors elementwise (with NumPy-style broadcasting support)

        Args:
            X: (non-differentiable) Tensor to prepend in concatenation

            Y: (non-differentiable) Tensor to append in concatenation
        """

        schema = get_schema("StringConcat", 20, "")
        op = Op(self, "StringConcat", schema)
        return op(*self._prepare_inputs(schema, X, Y))

    T1_StringSplit: TypeAlias = STRING

    T2_StringSplit: TypeAlias = STRING

    T3_StringSplit: TypeAlias = INT64

    def StringSplit(
        self,
        X: T1_StringSplit,
        *,
        delimiter: Optional[str] = None,
        maxsplit: Optional[int] = None,
    ) -> Tuple[T2_StringSplit, T3_StringSplit]:
        r"""[🌐 StringSplit(20)](https://onnx.ai/onnx/operators/onnx__StringSplit.html#stringsplit-20 "Online Documentation")

        StringSplit splits a string tensor's elements into substrings based on a delimiter attribute and a maxsplit attribute.

        The first output of this operator is a tensor of strings representing the substrings from splitting each input string on the `delimiter` substring. This tensor has one additional rank compared to the input tensor in order to store the substrings for each input element (where the input tensor is not empty). Note that, in order to ensure the same number of elements are present in the final dimension, this tensor will pad empty strings as illustrated in the examples below. Consecutive delimiters are not grouped together and are deemed to delimit empty strings, except if the `delimiter` is unspecified or is the empty string (""). In the case where the `delimiter` is unspecified or the empty string, consecutive whitespace characters are regarded as a single separator and leading or trailing whitespace is removed in the output.

        The second output tensor represents the number of substrings generated. `maxsplit` can be used to limit the number of splits performed - after the `maxsplit`th split if the string is not fully split, the trailing suffix of input string after the final split point is also added. For elements where fewer splits are possible than specified in `maxsplit`, it has no effect.

        Args:
            X: (non-differentiable) Tensor of strings to split.

            delimiter: Delimiter to split on. If left unset or set to the empty string
                (""), the input is split on consecutive whitespace.

            maxsplit: Maximum number of splits (from left to right). If left unset (or
                if the number of possible splits are less than maxsplit), it will make
                as many splits as possible. Note that the maximum possible number of
                substrings returned with `maxsplit` specified is `maxsplit+1` since the
                remaining suffix after the `maxsplit`th split is included in the output.
        """

        schema = get_schema("StringSplit", 20, "")
        op = Op(self, "StringSplit", schema)
        return op(*self._prepare_inputs(schema, X), delimiter=delimiter, maxsplit=maxsplit)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""
A one-layer Whisper encoder model test case, with inputs: audio_features.
This is an onnxscript version of the model.
"""

import numpy as np
import onnx_ir as ir

from onnxscript import script
from onnxscript.onnx_opset import opset18
from onnxscript.onnx_types import FLOAT


def make_model(
    encoder_encoder_embed_positions_weight,
    encoder_encoder_conv1_weight,
    encoder_encoder_conv1_bias,
    encoder_encoder_conv2_weight,
    encoder_encoder_conv2_bias,
    encoder_encoder_layers_0_self_attn_layer_norm_weight,
    encoder_encoder_layers_0_self_attn_layer_norm_bias,
    encoder_encoder_layers_0_self_attn_q_proj_weight,
    encoder_encoder_layers_0_self_attn_q_proj_bias,
    encoder_encoder_layers_0_self_attn_k_proj_weight,
    encoder_encoder_layers_0_self_attn_v_proj_weight,
    encoder_encoder_layers_0_self_attn_v_proj_bias,
    encoder_encoder_layers_0_self_attn_out_proj_weight,
    encoder_encoder_layers_0_self_attn_out_proj_bias,
    encoder_encoder_layers_0_final_layer_norm_weight,
    encoder_encoder_layers_0_final_layer_norm_bias,
    encoder_encoder_layers_0_fc1_weight,
    encoder_encoder_layers_0_fc1_bias,
    encoder_encoder_layers_0_fc2_weight,
    encoder_encoder_layers_0_fc2_bias,
    encoder_encoder_layer_norm_weight,
    encoder_encoder_layer_norm_bias,
):
    @script()
    def main_graph(
        audio_features: FLOAT[1, 80, 3000],
    ) -> FLOAT[1, 1500, 384]:
        val_0 = opset18.Shape(audio_features, end=1, start=0)
        conv1d = opset18.Conv(
            audio_features,
            encoder_encoder_conv1_weight,
            encoder_encoder_conv1_bias,
            group=1,
            pads=[1, 1],
            auto_pad="NOTSET",
            strides=[1],
            dilations=[1],
        )
        val_2 = opset18.Div(conv1d, 1.4142135)
        val_3 = opset18.Erf(val_2)
        val_5 = opset18.Add(val_3, 1.0)
        val_7 = opset18.Mul(0.5, val_5)
        gelu = opset18.Mul(conv1d, val_7)
        conv1d_1 = opset18.Conv(
            gelu,
            encoder_encoder_conv2_weight,
            encoder_encoder_conv2_bias,
            group=1,
            pads=[1, 1],
            auto_pad="NOTSET",
            strides=[2],
            dilations=[1],
        )
        val_9 = opset18.Div(conv1d_1, 1.4142135)
        val_10 = opset18.Erf(val_9)
        val_12 = opset18.Add(val_10, 1.0)
        val_14 = opset18.Mul(0.5, val_12)
        gelu_1 = opset18.Mul(conv1d_1, val_14)
        permute = opset18.Transpose(gelu_1, perm=[0, 2, 1])
        add_20 = opset18.Add(permute, encoder_encoder_embed_positions_weight)
        layer_norm = opset18.LayerNormalization(
            add_20,
            encoder_encoder_layers_0_self_attn_layer_norm_weight,
            encoder_encoder_layers_0_self_attn_layer_norm_bias,
            stash_type=1,
            epsilon=9.999999747378752e-06,
            axis=-1,
        )
        val_17 = opset18.Transpose(
            encoder_encoder_layers_0_self_attn_q_proj_weight, perm=[1, 0]
        )
        val_18 = opset18.MatMul(layer_norm, val_17)
        linear = opset18.Add(val_18, encoder_encoder_layers_0_self_attn_q_proj_bias)
        mul_18 = opset18.Mul(linear, 0.125)
        val_25 = opset18.Concat(val_0, [1500], [6], [64], axis=0)
        view = opset18.Reshape(mul_18, val_25, allowzero=0)
        transpose = opset18.Transpose(view, perm=[0, 2, 1, 3])
        val_27 = opset18.Transpose(
            encoder_encoder_layers_0_self_attn_k_proj_weight, perm=[1, 0]
        )
        linear_1 = opset18.MatMul(layer_norm, val_27)
        val_31 = opset18.Concat(val_0, [-1], [6], [64], axis=0)
        view_1 = opset18.Reshape(linear_1, val_31, allowzero=0)
        val_33 = opset18.Transpose(
            encoder_encoder_layers_0_self_attn_v_proj_weight, perm=[1, 0]
        )
        val_34 = opset18.MatMul(layer_norm, val_33)
        linear_2 = opset18.Add(val_34, encoder_encoder_layers_0_self_attn_v_proj_bias)
        val_37 = opset18.Concat(val_0, [-1], [6], [64], axis=0)
        view_2 = opset18.Reshape(linear_2, val_37, allowzero=0)
        transpose_2 = opset18.Transpose(view_2, perm=[0, 2, 1, 3])
        transpose_3 = opset18.Transpose(view_1, perm=[0, 2, 3, 1])
        matmul = opset18.MatMul(transpose, transpose_3)
        softmax = opset18.Softmax(matmul, axis=-1)
        matmul_1 = opset18.MatMul(softmax, transpose_2)
        transpose_4 = opset18.Transpose(matmul_1, perm=[0, 2, 1, 3])
        val_42 = opset18.Concat(val_0, [1500], [384], axis=0)
        _unsafe_view = opset18.Reshape(transpose_4, val_42, allowzero=0)
        val_44 = opset18.Transpose(
            encoder_encoder_layers_0_self_attn_out_proj_weight, perm=[1, 0]
        )
        val_45 = opset18.MatMul(_unsafe_view, val_44)
        linear_3 = opset18.Add(val_45, encoder_encoder_layers_0_self_attn_out_proj_bias)
        add_141 = opset18.Add(add_20, linear_3)
        layer_norm_1 = opset18.LayerNormalization(
            add_141,
            encoder_encoder_layers_0_final_layer_norm_weight,
            encoder_encoder_layers_0_final_layer_norm_bias,
            stash_type=1,
            epsilon=9.999999747378752e-06,
            axis=-1,
        )
        val_48 = opset18.Transpose(encoder_encoder_layers_0_fc1_weight, perm=[1, 0])
        val_49 = opset18.MatMul(layer_norm_1, val_48)
        linear_4 = opset18.Add(val_49, encoder_encoder_layers_0_fc1_bias)
        val_51 = opset18.Div(linear_4, 1.4142135)
        val_52 = opset18.Erf(val_51)
        val_54 = opset18.Add(val_52, 1.0)
        val_56 = opset18.Mul(0.5, val_54)
        gelu_2 = opset18.Mul(linear_4, val_56)
        val_57 = opset18.Transpose(encoder_encoder_layers_0_fc2_weight, perm=[1, 0])
        val_58 = opset18.MatMul(gelu_2, val_57)
        linear_5 = opset18.Add(val_58, encoder_encoder_layers_0_fc2_bias)
        add_170 = opset18.Add(add_141, linear_5)
        layer_norm_2 = opset18.LayerNormalization(
            add_170,
            encoder_encoder_layer_norm_weight,
            encoder_encoder_layer_norm_bias,
            stash_type=1,
            epsilon=9.999999747378752e-06,
            axis=-1,
        )
        return layer_norm_2

    model = main_graph.to_model_proto()
    return model


def make_model_with_random_weights():
    np.random.seed(10)  # Set a fixed seed
    encoder_encoder_embed_positions_weight = np.random.rand(1500, 384).astype(np.float32)
    encoder_encoder_conv1_weight = np.random.rand(384, 80, 3).astype(np.float32)
    encoder_encoder_conv1_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_conv2_weight = np.random.rand(384, 384, 3).astype(np.float32)
    encoder_encoder_conv2_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_self_attn_layer_norm_weight = np.random.rand(384).astype(
        np.float32
    )
    encoder_encoder_layers_0_self_attn_layer_norm_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_self_attn_q_proj_weight = np.random.rand(384, 384).astype(
        np.float32
    )
    encoder_encoder_layers_0_self_attn_q_proj_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_self_attn_k_proj_weight = np.random.rand(384, 384).astype(
        np.float32
    )
    encoder_encoder_layers_0_self_attn_v_proj_weight = np.random.rand(384, 384).astype(
        np.float32
    )
    encoder_encoder_layers_0_self_attn_v_proj_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_self_attn_out_proj_weight = np.random.rand(384, 384).astype(
        np.float32
    )
    encoder_encoder_layers_0_self_attn_out_proj_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_final_layer_norm_weight = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_final_layer_norm_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layers_0_fc1_weight = np.random.rand(1536, 384).astype(np.float32)
    encoder_encoder_layers_0_fc1_bias = np.random.rand(1536).astype(np.float32)
    encoder_encoder_layers_0_fc2_weight = np.random.rand(384, 1536).astype(np.float32)
    encoder_encoder_layers_0_fc2_bias = np.random.rand(384).astype(np.float32)
    encoder_encoder_layer_norm_weight = np.random.rand(384).astype(np.float32)
    encoder_encoder_layer_norm_bias = np.random.rand(384).astype(np.float32)
    model = make_model(
        encoder_encoder_embed_positions_weight,
        encoder_encoder_conv1_weight,
        encoder_encoder_conv1_bias,
        encoder_encoder_conv2_weight,
        encoder_encoder_conv2_bias,
        encoder_encoder_layers_0_self_attn_layer_norm_weight,
        encoder_encoder_layers_0_self_attn_layer_norm_bias,
        encoder_encoder_layers_0_self_attn_q_proj_weight,
        encoder_encoder_layers_0_self_attn_q_proj_bias,
        encoder_encoder_layers_0_self_attn_k_proj_weight,
        encoder_encoder_layers_0_self_attn_v_proj_weight,
        encoder_encoder_layers_0_self_attn_v_proj_bias,
        encoder_encoder_layers_0_self_attn_out_proj_weight,
        encoder_encoder_layers_0_self_attn_out_proj_bias,
        encoder_encoder_layers_0_final_layer_norm_weight,
        encoder_encoder_layers_0_final_layer_norm_bias,
        encoder_encoder_layers_0_fc1_weight,
        encoder_encoder_layers_0_fc1_bias,
        encoder_encoder_layers_0_fc2_weight,
        encoder_encoder_layers_0_fc2_bias,
        encoder_encoder_layer_norm_weight,
        encoder_encoder_layer_norm_bias,
    )
    return model


class _WhisperEncoderTest:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = make_model_with_random_weights()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            np.random.seed(10)  # Set a fixed seed
            inputs = {
                "audio_features": np.random.rand(1, 80, 3000).astype(np.float32),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def whisper_encoder_test():
    return _WhisperEncoderTest()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import math

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

"""
Multi-Head Attention (MHA) pre-scaling fusion patterns.

This module contains rewrite rules for fusing scale operations that occur before
Multi-Head Attention operations. The fusion optimizes patterns where a query tensor
is scaled before being passed to MHA by incorporating the scaling directly into
the MHA operation.

Example pattern:
    query -> Mul(scale) -> MultiHeadAttention -> output

Gets rewritten to:
    query -> MultiHeadAttention(with integrated scaling) -> output
"""


class FuseMHAScale(pattern.RewriteRuleClassBase):
    def pattern(self, op, query, scale):
        scaled_query = op.Mul(query, scale)
        mha_output = op.MultiHeadAttention(
            scaled_query,
            _allow_other_inputs=True,
            _domain="com.microsoft",
            _outputs=["mha_output"],
        )
        return mha_output

    def check(self, context, scale, **_):
        scale_value = _ir_utils.get_singleton_value(scale)
        if scale_value is None or not isinstance(scale_value, (int, float)):
            return pattern.MatchResult().fail("Scale must be a constant numeric value.", scale)
        self._scale = scale_value
        return True

    def rewrite(self, op, query, mha_output, **_):
        # Integrate the scale into the MHA operation
        mha_node = mha_output.producer()
        assert mha_node is not None
        # Compute original scale factor for MHA:
        attributes = mha_node.attributes
        original_scale = attributes.get_float("scale", None)
        if original_scale is None:
            num_heads = attributes.get_int("num_heads", None)
            if num_heads is None:
                return None
            head_size = query.shape[-1] // num_heads
            original_scale = 1.0 / math.sqrt(head_size)
        self._scale *= original_scale
        inputs = list(mha_node.inputs)
        inputs[0] = query
        attributes = dict(attributes)
        attributes["scale"] = self._scale
        return op.MultiHeadAttention(
            *inputs, **attributes, _domain="com.microsoft", _outputs=1
        )


_mha_scale_rules = pattern.RewriteRuleSet([FuseMHAScale.rule()])

fuse_mha_scale = _fusion_utils.apply_fusion_rules(_mha_scale_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Any, Optional, Sequence

import numpy as np
import onnx
from onnx import FunctionProto, GraphProto, ModelProto, TensorProto, ValueInfoProto

import onnxscript.onnx_types
from onnxscript._internal import type_annotation

_SINGLE_INDENT = "    "

_SMALL_TENSOR_SIZE = 4

kwlist = {
    "False",
    "None",
    "True",
    "and",
    "as",
    "assert",
    "async",
    "await",
    "break",
    "class",
    "continue",
    "def",
    "del",
    "elif",
    "else",
    "except",
    "finally",
    "for",
    "from",
    "global",
    "if",
    "import",
    "in",
    "is",
    "lambda",
    "nonlocal",
    "not",
    "or",
    "pass",
    "raise",
    "return",
    "try",
    "while",
    "with",
    "yield",
}


def _get_const_repr(const_node):
    """Given an ONNX Constant-op node, returns a string representation of
    the constant-value in ONNXScript, if a compact representation is possible.
    Returns None otherwise.
    Supports only FLOAT/INT64 values and scalars and small rank-1 tensors.
    This needs to be reconciled with the ONNXScript converter.
    """
    assert const_node.op_type == "Constant", "Expected a constant node"
    attr = const_node.attribute[0]
    if not attr.HasField("t"):
        return None
    tensor_proto = attr.t
    if tensor_proto.data_type in {TensorProto.FLOAT, TensorProto.INT64}:
        rank = len(tensor_proto.dims)
        if rank == 0:
            array = onnx.numpy_helper.to_array(tensor_proto).reshape(1)  # noqa: TID251
            return str(array[0])
        if rank == 1 and tensor_proto.dims[0] < 5:
            nparray = onnx.numpy_helper.to_array(tensor_proto)  # noqa: TID251
            return repr(nparray.tolist())
    return None


def _cleanup_variable_name(name: ValueInfoProto | str) -> str:
    """Converts given name into a valid python variable names.
    Handles names that clash with python keywords and common issues seen in ONNX models:
    * Identifiers like "5" (that do not start with an alpha character)
    * Identifiers that contain a dot like "layers.0.foo"
    This is a simple heuristic, and doesn't guarantee it avoids name-clashes.
    """
    if isinstance(name, ValueInfoProto):
        # Handle graph/function input/output uniformly
        name = name.name
    assert isinstance(name, str)
    assert name != ""
    if name in kwlist:
        return f"r_{name}"
    first = name[0]
    if not (first.isalpha() or (first == "_")):
        name = f"__{name}"

    def rename_char(char):
        """Replace invalid character by underscore."""
        return char if (char.isalnum() or (char == "_")) else "_"

    return "".join([rename_char(c) for c in name])


def _make_short_name_mapper():
    """Returns a renamer used to create short new names  (like v0, v1, ...) for variables."""
    variable_names: dict[str, str] = {}

    def renamer(name):
        # TODO: simplify this. No need to use _cleanup_variable_name?
        var_name = _cleanup_variable_name(name)
        if var_name in variable_names:
            return variable_names[var_name]
        new_name = f"v{len(variable_names) + 1}"
        assert var_name is not None  # TODO(rama): This looks suspect.
        variable_names[var_name] = new_name
        return new_name

    return renamer


def _translate_type(onnx_type):
    """Converts a onnx type into a type defined by *onnxscript*."""
    return onnxscript.onnx_types.onnx_type_to_onnxscript_repr(onnx_type, reversible=False)


def _translate_signature(inputs, outputs):
    """Produce the script-functions signature."""

    def input_sig(inp: ValueInfoProto | str):
        if isinstance(inp, ValueInfoProto):
            # GraphProto inputs/outputs are ValueInfoProto
            return f"{_cleanup_variable_name(inp.name)}: {_translate_type(inp.type)}"

        # FunctionProto inputs/outputs are just strings
        return _cleanup_variable_name(inp)

    result = f"({', '.join([input_sig(x) for x in inputs])})"
    if outputs and isinstance(outputs[0], ValueInfoProto):
        result += f" -> ({', '.join([_translate_type(x.type) for x in outputs])})"
    return f"{result}:"


def _translate_value_infos(value_infos: Sequence[ValueInfoProto]) -> str:
    def _translate_value_info(value_info: ValueInfoProto) -> str:
        return f"{_SINGLE_INDENT}'{_cleanup_variable_name(value_info.name)}': {_translate_type(value_info.type)},"

    lines = [_translate_value_info(x) for x in value_infos]
    lines_joined = "\n".join(lines)
    return "{\n" + lines_joined + "\n}"


def _to_str(s):
    if isinstance(s, bytes):
        try:
            return s.decode("utf-8")
        except UnicodeDecodeError:
            pass
    return s


def _is_attribute_ref(attr: onnx.AttributeProto) -> bool:
    return attr.HasField("ref_attr_name") and attr.ref_attr_name != ""


def _attribute_value(attr: onnx.AttributeProto):
    if attr.type == onnx.AttributeProto.FLOAT:
        return attr.f
    if attr.type == onnx.AttributeProto.INT:
        return attr.i
    if attr.type == onnx.AttributeProto.STRING:
        return _to_str(attr.s)
    if attr.type == onnx.AttributeProto.TENSOR:
        tensor_proto = attr.t
        if onnx.external_data_helper.uses_external_data(tensor_proto):
            return tensor_proto
        else:
            return onnx.numpy_helper.to_array(tensor_proto)  # noqa: TID251
    # TODO:
    # - onnx.AttributeProto.GRAPH
    # - onnx.AttributeProto.SPARSE_TENSOR
    # - onnx.AttributeProto.TYPE_PROTO
    if attr.type == onnx.AttributeProto.FLOATS:
        return list(attr.floats)
    if attr.type == onnx.AttributeProto.INTS:
        return list(attr.ints)
    if attr.type == onnx.AttributeProto.STRINGS:
        return list(map(_to_str, attr.strings))
    # TODO:
    # - onnx.AttributeProto.TENSORS
    # - onnx.AttributeProto.GRAPHS
    # - onnx.AttributeProto.SPARSE_TENSORS
    # - onnx.AttributeProto.TYPE_PROTOS
    raise NotImplementedError(f"Unable to return a value for attribute {attr!r}.")


def _update_names_used_in_graph(names: set[str], graph: GraphProto) -> None:
    """Adds the names used in a graph to given set."""
    names.update(x.name for x in graph.input)
    names.update(x.name for x in graph.output)
    names.update(x.name for x in graph.initializer)
    for node in graph.node:
        _update_names_used_in_node(names, node)


def _update_names_used_in_node(names: set[str], node: onnx.NodeProto) -> None:
    names.update(node.input)
    names.update(node.output)
    for attr in node.attribute:
        if attr.HasField("g"):
            _update_names_used_in_graph(names, attr.g)
        for g in attr.graphs:
            _update_names_used_in_graph(names, g)


def _update_names_used_in_function(names: set[str], fun: FunctionProto) -> None:
    names.update(fun.input)
    names.update(fun.output)
    for node in fun.node:
        _update_names_used_in_node(names, node)


def _names_used_in_function(fun: FunctionProto) -> set[str]:
    names: set[str] = set()
    _update_names_used_in_function(names, fun)
    return names


def has_input(node: onnx.NodeProto, index: int) -> bool:
    """Returns True iff the node has an input at given index."""
    return index < len(node.input) and node.input[index] != ""


def is_onnx_op(node: onnx.NodeProto, op_type: str) -> bool:
    return node.op_type == op_type and node.domain in {"", "ai.onnx"}


def _is_used_in_graph_body(name: str, graph: GraphProto) -> bool:
    """Returns True iff the given name is used in the graph body."""
    # TODO: This is an approximation.
    names: set[str] = set()
    for node in graph.node:
        _update_names_used_in_node(names, node)
    return name in names


def _cond_is_used_in_loop_body(graph: GraphProto) -> bool:
    """Returns True iff loop requires a condition."""
    cond_in = graph.input[1].name
    cond_out = graph.output[0].name
    for node in graph.node:
        # Ignore "cond_out = Identity(cond_in)" node
        if is_onnx_op(node, "Identity") and len(node.input) == 1 and len(node.output) == 1:
            if node.input[0] == cond_in and node.output[0] == cond_out:
                continue
        names: set[str] = set()
        # TODO: The following is an approximation-based check
        _update_names_used_in_node(names, node)
        if (cond_in in names) or (cond_out in names):
            return True
    return False


class _Exporter:
    """Class used for recursive traversal of Proto structures."""

    def __init__(
        self, *, rename: bool, use_operators: bool, inline_const: bool, skip_initializers: bool
    ) -> None:
        self.use_operators = use_operators
        if rename:
            rename_function = _make_short_name_mapper()
        else:
            rename_function = _cleanup_variable_name
        self._rename_variable = self._handle_attrname_conflict(rename_function)
        self.inline_const = inline_const
        self.constants: dict[str, str] = {}
        self._attr_renaming: dict[str, str | None] = {}  # For current function.
        self._names_used: set[str] = set()  # For current function.
        # _name_remappings: used to undo the SSA-renaming in ONNX control-flow ops.
        # We map the multiple SSA-variants back to the same Python variable name.
        self._name_remappings: list[dict[str, str]] = []
        self.skip_initializers = skip_initializers
        self.skipped_initializers: dict[str, onnx.TensorProto] = {}

    def _handle_attrname_conflict(self, renamer):
        """Add ref-attr-name-conflict handling logic to renaming function."""

        def new_renamer(name):
            new_name = renamer(name)
            if new_name not in self._attr_renaming:
                return new_name
            # Name conflicts with attribute parameter name.
            alternate = self._attr_renaming[new_name]
            if alternate is not None:
                return alternate
            counter = 0
            candidate = new_name
            while candidate in self._names_used:
                candidate = f"{new_name}_{counter}"
                counter += 1
            self._attr_renaming[new_name] = candidate
            self._names_used.add(candidate)
            return candidate

        return new_renamer

    def _translate_onnx_var(self, var):
        """Converts an ONNX variable name to a python variable name."""
        if isinstance(var, ValueInfoProto):
            var = var.name
        if var == "":
            return "None"
        for scope in reversed(self._name_remappings):
            if var in scope:
                return scope[var]
        return self._rename_variable(var)

    def _translate_onnx_var_ref(self, var):
        """Translates a reference to an ONNX variable (a r-value)"""
        if var in self.constants:
            return self.constants[var]

        return self._translate_onnx_var(var)

    def _rename_domain(self, domain: str) -> str:
        if domain in {"", "ai.onnx"}:
            return "opset"  # TODO: Need checks to avoid name conflicts.
        return _cleanup_variable_name(domain)  # type: ignore[return-value]

    def _make_opset_name(self, domain, version):
        return f"{self._rename_domain(domain)}{version}"

    def _make_callee_name(self, domain, version, name, node=False):
        """Generate name to be used for called op/function in a node or for a generated script function."""
        # TODO: Avoid name-conflict between function-names and value-names
        name = _cleanup_variable_name(name)
        if node:
            if version is None:
                version = 1
            if not isinstance(version, int):
                raise TypeError(
                    f"version must be an integer not {version!r} for domain={domain!r} "
                    f"and name={name!r}."
                )
            opset = self._make_opset_name(domain, version)
            return f"{opset}.{name}"
        return name

    def _translate_graph_body(self, graph, opsets, indent=0):
        """Translates a graph body into python.
        The graph may be the main graph (of a model) or a subgraph (of a Loop or If node).
        """
        code = []
        if hasattr(graph, "initializer"):
            for init in graph.initializer:
                if self.skip_initializers:
                    size = 1
                    for d in init.dims:
                        size *= d
                    if size > _SMALL_TENSOR_SIZE:
                        init_py_name = self._translate_onnx_var(init.name)
                        if init_py_name in self.skipped_initializers:
                            raise RuntimeError(
                                f"Initializer {init.name!r} is already present in skipped_initializers."
                            )
                        self.skipped_initializers[init_py_name] = init
                        continue
                node = onnx.helper.make_node(  # noqa: TID251
                    "Constant",
                    [],
                    [self._translate_onnx_var(init.name)],  # type: ignore[list-item]
                    value=init,
                )
                pyinit = self._translate_node(node, opsets, indent=indent)
                if pyinit:
                    code.append(pyinit)
        if hasattr(graph, "sparse_initializer") and len(graph.sparse_initializer) > 0:
            raise NotImplementedError("Unable to convert sparse_initilizer into python.")
        for node in graph.node:
            pynode = self._translate_node(node, opsets, indent=indent)
            if pynode:
                if node.name:
                    pynode += f"  # {node.name}"
                code.append(pynode)

        final = "\n".join(code)
        return final

    def _translate_attributes(self, node):
        attributes = []
        for at in node.attribute:
            if _is_attribute_ref(at):
                attributes.append((at.name, at.ref_attr_name))
                continue
            value = _attribute_value(at)
            if isinstance(value, (str, bytes)):
                attributes.append((at.name, repr(value)))
                continue
            if isinstance(value, np.ndarray):
                onnx_dtype = at.t.data_type
                if len(value.shape) == 0:
                    text = (
                        f'make_tensor("value", {onnx_dtype}, dims=[], '
                        f"vals=[{repr(value.tolist()).replace('nan', 'np.nan').replace('inf', 'np.inf')}])"
                    )
                else:
                    text = (
                        f'make_tensor("value", {onnx_dtype}, dims={list(value.shape)!r}, '
                        f"vals={repr(value.ravel().tolist()).replace('nan', 'np.nan').replace('inf', 'np.inf')})"
                    )
                attributes.append((at.name, text))
                continue
            if isinstance(value, TensorProto):
                metadata = onnx.external_data_helper.ExternalDataInfo(value)
                name = value.name or "value"
                text = "external_tensor("
                text += f"{name!r}, {value.data_type}, {list(value.dims)!r}"
                text += f", {metadata.location!r}"
                if metadata.offset:
                    text += f", offset={metadata.offset!r}"
                if metadata.length:
                    text += f", length={metadata.length!r}"
                text += ")"
                attributes.append((at.name, text))
                continue
            attributes.append((at.name, repr(value)))

        return ", ".join(f"{k}={v}" for k, v in attributes)

    def _translate_if(self, node, opsets, indent=0):
        """Translates a node If into python."""
        sindent = _SINGLE_INDENT * indent
        cond = self._translate_onnx_var_ref(node.input[0])
        code = [f"{sindent}if {cond}:"]
        if len(node.attribute) != 2:
            raise RuntimeError(
                f"Node {node.op_type!r} expected two attributes not {len(node.attribute)}."
            )
        atts = node.attribute
        if atts[0].name == "else_branch":
            else_branch, then_branch = atts[0].g, atts[1].g
        else:
            else_branch, then_branch = atts[1].g, atts[0].g

        code.append(
            self._translate_graph_body(
                then_branch,
                opsets,
                indent=indent + 1,
            )
        )
        code.extend(self._emit_assign(node.output, then_branch.output, indent + 1))

        code.append(f"{sindent}else:")
        code.append(
            self._translate_graph_body(
                else_branch,
                opsets,
                indent=indent + 1,
            )
        )
        code.extend(self._emit_assign(node.output, else_branch.output, indent + 1))
        return "\n".join(code)

    def _emit_assign(self, lhs, rhs, indent):
        def to_var(x):
            if isinstance(x, ValueInfoProto):
                x = x.name
            return self._translate_onnx_var(x)

        sindent = _SINGLE_INDENT * indent

        def assign(lhs_var: str, rhs_var: str):
            return f"{sindent}{to_var(lhs_var)} = {to_var(rhs_var)}"

        if isinstance(lhs, (str, ValueInfoProto)):
            return [assign(lhs, rhs)]
        return [assign(x, y) for x, y in zip(lhs, rhs)]

    def _translate_loop(self, node, opsets, indent=0):
        """Translates a node Loop into python."""
        body = node.attribute[0].g
        sindent = _SINGLE_INDENT * indent
        rows = []

        # Node inputs: optional max-trip-count, optional condition, initial values (of dependencies)
        # Node outputs: final values (of dependencies), scan-outputs
        # Body inputs: iteration-count, condition, input values (of dependencies)
        # Body outputs: condition, output values (of dependencies), scan-outputs

        onnx_iter_var = body.input[0].name
        if has_input(node, 0):
            use_iter_var = True
            n_iter = self._translate_onnx_var(node.input[0])
        else:
            use_iter_var = _is_used_in_graph_body(onnx_iter_var, body)
            n_iter = None
        iter_var = self._translate_onnx_var(onnx_iter_var)

        cond_in = body.input[1].name
        cond_out = body.output[0].name
        py_cond = self._translate_onnx_var(cond_in)
        use_loop_cond = True  # TODO
        if has_input(node, 1):
            rows.extend(self._emit_assign(cond_in, node.input[1], indent))
        else:
            use_loop_cond = _cond_is_used_in_loop_body(body)
            # rows.append(f"{sindent}{py_cond} = True")

        num_state_vars = max(len(node.input) - 2, 0)
        actual_ins = node.input[2:]
        formal_ins = body.input[2:]
        formal_outs = body.output[1 : num_state_vars + 1]
        actual_outs = node.output[0:num_state_vars]

        rows.extend(self._emit_assign(formal_ins, actual_ins, indent))

        if node.name:
            node_name = "  # " + node.name
        else:
            node_name = ""
        if use_iter_var and not use_loop_cond:
            rows.append(f"{sindent}for {iter_var} in range({n_iter}):{node_name}")
            # The following is a hacky way to suppress the generation of
            # "cond_out = cond_in", which ONNX forces for a FOR loop.
            # TODO: a cleaner solution for this.
            self._name_remappings[-1][cond_out] = self._translate_onnx_var(cond_in)
        elif not use_iter_var and use_loop_cond:
            rows.append(f"{sindent}while {py_cond}:{node_name}")
        elif use_iter_var and use_loop_cond:
            # TODO: This needs fixing
            rows.append(f"{sindent}for {iter_var} in range({n_iter}):{node_name}")
            rows.append(f"{sindent}{_SINGLE_INDENT}if not {py_cond}:")
            rows.append(f"{sindent}{_SINGLE_INDENT * 2}break")
        else:
            raise RuntimeError(
                f"Unable to export loop type {node.op_type!r} into python because "
                "there is no stop condition."
            )

        rows.append(
            self._translate_graph_body(
                body,
                opsets,
                indent=indent + 1,
            )
        )
        if use_loop_cond:
            rows.extend(self._emit_assign(cond_in, cond_out, indent + 1))
        rows.extend(self._emit_assign(formal_ins, formal_outs, indent + 1))
        rows.extend(self._emit_assign(actual_outs, formal_ins, indent))

        # TODO: This doesn't handle scan-outputs yet.

        return "\n".join(rows)

    def _translate_scan(self, node, opsets, indent=0):
        """Translates a node Scan into python."""
        raise NotImplementedError()

    def _translate_node(self, onnx_node, opsets, indent=0):
        if isinstance(onnx_node, dict):
            node = onnx_node["onnx_node"]
        else:
            node = onnx_node
        if self.inline_const and node.op_type == "Constant":
            val = _get_const_repr(node)
            if val is not None:
                self.constants[node.output[0]] = str(val)
                return ""
        if node.op_type in {"If", "Loop", "Scan"}:
            # If, Loop, Scan
            if node.op_type == "If":
                return self._translate_if(node, opsets, indent=indent)
            if node.op_type == "Loop":
                return self._translate_loop(node, opsets, indent=indent)
            if node.op_type == "Scan":
                return self._translate_scan(node, opsets, indent=indent)
            raise RuntimeError(f"Unable to export node type {node.op_type!r} into python.")
        if any(hasattr(att, "g") and att.g and att.g.ByteSize() > 0 for att in node.attribute):
            raise RuntimeError(f"Unable to export node type {node.op_type!r} into python.")
        ops = {
            "Add": "+",
            "Sub": "-",
            "Mul": "*",
            "MatMul": "@",
            "Div": "/",
            "Pow": "**",
            "And": "&",
            "Or": "|",
            "Greater": ">",
            "Equal": "==",
            "Lesser": "<",
            "GreaterOrEqual": ">=",
            "LessOrEqual": "<=",
        }
        sindent = _SINGLE_INDENT * indent
        if self.use_operators and node.op_type in ops:
            return (
                f"{sindent}{self._translate_onnx_var(node.output[0])} = "
                f"{(f' {ops[node.op_type]} ').join(map(self._translate_onnx_var_ref, node.input))}"
            )
        callee_name = self._make_callee_name(
            node.domain, opsets[node.domain], node.op_type, node=True
        )
        attributes_str = self._translate_attributes(node)
        if len(node.input) > 0 and len(attributes_str) > 0:
            attributes_str = f", {attributes_str}"
        output_names: list[Any] = []
        for i, o in enumerate(node.output):
            if o in ("", None):
                output_names.append(f"_{i}")
            else:
                output_names.append(self._translate_onnx_var(o))

        input_names = [self._translate_onnx_var_ref(x) for x in node.input]

        # Suppress generation of redundant copy: used to suppress "cond_out = cond_in"
        # from an ONNX FOR loop, which can cause problems in python.
        if node.op_type == "Identity" and len(node.input) == 1 and len(node.output) == 1:
            if output_names[0] == input_names[0]:
                return ""
        text = [
            sindent,
            ", ".join(output_names),
            " = ",
            callee_name,
            "(",
            ", ".join(input_names),
            attributes_str,
            ")",
        ]
        return "".join(text)

    def _translate_opset_import(self, domain: str, version: int) -> str:
        if domain in {"", "ai.onnx"}:
            return f"from onnxscript.onnx_opset import opset{version}\n"
        else:
            varname = self._make_opset_name(domain, version)
            return f"{varname} = Opset('{domain}', {version})\n"

    def _translate_opset_imports(
        self, opset_imports: Sequence[onnx.OperatorSetIdProto]
    ) -> str:
        return "".join(
            [self._translate_opset_import(x.domain, x.version) for x in opset_imports]
        )

    def _translate_opset_imports_of(
        self, proto: ModelProto | FunctionProto | GraphProto
    ) -> str:
        if hasattr(proto, "opset_import"):
            text = self._translate_opset_imports(proto.opset_import)
            if isinstance(proto, FunctionProto):
                if not any(x.domain == proto.domain for x in proto.opset_import):
                    text += self._translate_opset_import(proto.domain, 1)
            return text
        return ""

    def _translate_function_signature(self, funproto: onnx.FunctionProto) -> str:
        """Generate signature for FunctionProto."""
        type_map = _attribute_param_types(funproto)

        def attr_sig(attr_name: str) -> str:
            self._attr_renaming[attr_name] = None
            self._names_used.add(attr_name)
            # A default type of INT is used for attribute parameters that are never used.
            type = type_map.get(attr_name, onnx.AttributeProto.INT)
            typerep = type_annotation.onnx_attr_type_to_onnxscript_repr(type)
            return f"{attr_name}: {typerep}"

        inputs = [self._translate_onnx_var(x) for x in funproto.input]
        attrs = [attr_sig(x) for x in funproto.attribute]
        input_and_attrs = ", ".join(inputs + attrs)  # type: ignore[arg-type]
        if len(funproto.attribute_proto) > 0:
            message = "\n   # Attribute parameters default-values not handled yet."
        else:
            message = ""
        return f"({input_and_attrs}):{message}"

    def _translate_function(self, funproto: onnx.FunctionProto) -> str:
        """Generate python code for FunctionProto."""
        opsets = {}
        for imported in funproto.opset_import:
            opsets[imported.domain] = imported.version
        self._attr_renaming = {}
        used_proto_names = _names_used_in_function(funproto)
        renamed_names_used = [self._translate_onnx_var(x) for x in used_proto_names]
        self._names_used = set(renamed_names_used)
        result = []

        def add_line(line: str) -> None:
            result.append(line)

        opset_name = self._make_opset_name(funproto.domain, 1)
        add_line(f"@script({opset_name})")
        fun_name = self._make_callee_name(funproto.domain, 1, funproto.name)
        fun_sig = self._translate_function_signature(funproto)
        add_line(f"def {fun_name}{fun_sig}")
        if funproto.doc_string:
            add_line(f'    """{funproto.doc_string}"""')
        self._name_remappings.append({})
        for node in funproto.node:
            add_line(self._translate_node(node, opsets, indent=1))
        return_values = ", ".join(self._translate_onnx_var(x) for x in funproto.output)
        add_line(f"    return {return_values}")
        self._name_remappings.pop()
        return "\n".join(result)

    def _translate_graph(self, model: onnx.ModelProto, function_name: Optional[str]) -> str:
        graph = model.graph
        opsets = {}
        for imported in model.opset_import:
            opsets[imported.domain] = imported.version
        if function_name is None:
            function_name = _cleanup_variable_name(graph.name)

        result: list[str] = []

        def add(line: str) -> None:
            result.append(line)

        if self.skip_initializers:
            indent_level = 2
            indent = _SINGLE_INDENT
        else:
            indent_level = 1
            indent = ""
        add(f"{indent}@script()")
        add(f"{indent}def {function_name}{_translate_signature(graph.input, graph.output)}")
        indent = indent + _SINGLE_INDENT
        doc = graph.doc_string
        if doc:
            add(f'{indent}"""{doc}"""')
        add(self._translate_graph_body(graph, opsets, indent=indent_level))
        return_values = ", ".join(self._translate_onnx_var(x) for x in graph.output)
        add(f"{indent}return {return_values}")
        script = "\n".join(result)
        if self.skipped_initializers:
            value_infos = _translate_value_infos(graph.value_info)
            return self._substitute_initializers(script, function_name, value_infos)
        return script

    def _substitute_initializers(
        self, script: str, script_function_name: str, value_infos: str
    ) -> str:
        init_names = self.skipped_initializers.keys()
        # Formal parameters representing initializers (single level indentation)
        __ = _SINGLE_INDENT
        initializers_as_params = "\n".join(f"{__}{x}," for x in init_names)

        def generate_rand(name: str, value: TensorProto) -> str:
            shape = ",".join(str(d) for d in value.dims)
            if value.data_type == TensorProto.FLOAT:
                return f"{__}{name} = np.random.rand({shape}).astype(np.float32)"
            if value.data_type == TensorProto.INT8:
                return f"{__}{name} = np.random.randint(-128, 127, size=({shape},), dtype=np.int8)"
            raise NotImplementedError(
                f"Unable to generate random initializer for data type {value.data_type}."
            )

        random_initializer_values = "\n".join(
            generate_rand(key, value) for key, value in self.skipped_initializers.items()
        )
        # Actual parameter values for initializers (double level indentation)
        indented_initializers_as_params = "\n".join(f"{__}{__}{x}," for x in init_names)
        return f"""
value_infos = {value_infos}

def make_model(
{initializers_as_params}
):
{script}

{__}model = {script_function_name}.to_model_proto(value_infos=value_infos)
{__}return model

def make_model_with_random_weights():
{random_initializer_values}
{__}model = make_model(
{indented_initializers_as_params}
{__})
{__}return model
"""

    def _import_onnx_types(
        self, proto: onnx.ModelProto | onnx.GraphProto | onnx.FunctionProto
    ) -> str:
        """Generate import statements for types used in the graph."""
        if isinstance(proto, ModelProto):
            graph_or_function = proto.graph
        else:
            graph_or_function = proto
        used_types: set[str] = set()
        for t in list(graph_or_function.input) + list(graph_or_function.output):
            if hasattr(t, "type"):
                ts = _translate_type(t.type)
                its = ts.split("[", maxsplit=1)[0]
                used_types.add(its)
        # TODO: handle types in nested graphs.
        sorted_types = sorted(used_types)
        if sorted_types:
            return "from onnxscript.onnx_types import " + ", ".join(sorted_types)
        return ""

    def export(
        self, proto: onnx.ModelProto | onnx.FunctionProto, function_name: Optional[str]
    ) -> str:
        result: list[str] = []

        def add(line: str) -> None:
            result.append(line)

        # Generic imports.
        add("import numpy as np")
        add("from onnx import TensorProto")
        add("from onnx.helper import make_tensor")
        add("from onnxscript import script, external_tensor")
        add("from onnxscript.values import Opset")
        add(self._import_onnx_types(proto))

        if isinstance(proto, ModelProto):
            translated_functions = [self._translate_function(f) for f in proto.functions]
            translated_functions.append(self._translate_graph(proto, function_name))
        else:
            assert isinstance(proto, FunctionProto)
            translated_functions = [self._translate_function(proto)]

        # TODO: unique_function_domain_version.add((f.domain, 1))
        add(self._translate_opset_imports_of(proto))
        result.extend(translated_functions)

        add("")
        final = "\n".join(result)

        if "\nreturn" in final:
            raise SyntaxError(f"The produced code is wrong.\n{final}")

        return final


def _attribute_param_types(
    funproto: onnx.FunctionProto,
) -> dict[str, onnx.AttributeProto.AttributeType]:
    """Compute mapping from (names of) attribute parameters of function to their types."""
    type_map = {}

    def visit_node(node: onnx.NodeProto) -> None:
        for attr in node.attribute:
            if _is_attribute_ref(attr):
                type_map[attr.ref_attr_name] = attr.type
            elif attr.type == onnx.AttributeProto.GRAPH:
                visit_graph(attr.g)
            elif attr.type == onnx.AttributeProto.GRAPHS:
                for graph in attr.graphs:
                    visit_graph(graph)

    def visit_graph(graph: onnx.GraphProto) -> None:
        for node in graph.node:
            visit_node(node)

    for node in funproto.node:
        visit_node(node)
    return type_map


def export2python(
    model_onnx,
    function_name: Optional[str] = None,
    *,
    rename: bool = False,
    use_operators: bool = False,
    inline_const: bool = False,
    skip_initializers: bool = False,
):
    """Exports an ONNX model to the *python* syntax.

    Args:
        model_onnx: string or ONNX graph
        rename: rename the names to get shorter names
        function_name: main function name
        use_operators: use Python operators.
        inline_const: replace ONNX constants inline if compact
        skip_initializers: generated script will not include initializers.
          Instead, a function that generates the model, given initializer values, is generated,
          along with one that generates random values for the initializers.

    Returns:
        python code
    The following example shows what a python code creating a graph
    implementing the KMeans would look like.
    .. runpython::
        :showcode:
        :process:
        import numpy as np
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2python
        X = np.arange(20).reshape(10, 2).astype(np.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)
        onx = to_onnx(tr, X, target_opset=14)
        code = export2python(onx)
        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, (ModelProto, FunctionProto)):
        raise TypeError(f"The function expects a ModelProto not {type(model_onnx)!r}.")

    exporter = _Exporter(
        rename=rename,
        use_operators=use_operators,
        inline_const=inline_const,
        skip_initializers=skip_initializers,
    )
    return exporter.export(model_onnx, function_name)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

__all__ = [
    "script",
    "graph",
    "ir",
    "nn",
    "optimizer",
    "rewriter",
    "version_converter",
    "export_onnx_lib",
    "OnnxFunction",
    "TracedOnnxFunction",
    "GraphBuilder",
    "OpBuilder",
    "proto2python",
    "external_tensor",
    "BFLOAT16",
    "FLOAT16",
    "FLOAT8E4M3FN",
    "FLOAT8E4M3FNUZ",
    "FLOAT8E5M2",
    "FLOAT8E5M2FNUZ",
    "FLOAT",
    "DOUBLE",
    "INT8",
    "INT16",
    "INT32",
    "INT64",
    "UINT8",
    "UINT16",
    "UINT32",
    "UINT64",
    "BOOL",
    "STRING",
    "COMPLEX64",
    "COMPLEX128",
    "opset1",
    "opset2",
    "opset3",
    "opset4",
    "opset5",
    "opset6",
    "opset7",
    "opset8",
    "opset9",
    "opset10",
    "opset11",
    "opset12",
    "opset13",
    "opset14",
    "opset15",
    "opset16",
    "opset17",
    "opset18",
    "opset19",
    "opset20",
    "opset21",
    "opset22",
    "opset23",
    "opset_ai_onnx_ml1",
    "opset_ai_onnx_ml2",
    "opset_ai_onnx_ml3",
    "opset_ai_onnx_ml4",
    "opset_ai_onnx_ml5",
    "DEBUG",
]

import importlib.metadata

from ._internal.main import export_onnx_lib, graph, script
from .backend.onnx_export import export2python as proto2python

# isort: off
from .onnx_opset import (
    opset1,
    opset2,
    opset3,
    opset4,
    opset5,
    opset6,
    opset7,
    opset8,
    opset9,
    opset10,
    opset11,
    opset12,
    opset13,
    opset14,
    opset15,
    opset16,
    opset17,
    opset18,
    opset19,
    opset20,
    opset21,
    opset22,
    opset23,
    opset_ai_onnx_ml1,
    opset_ai_onnx_ml2,
    opset_ai_onnx_ml3,
    opset_ai_onnx_ml4,
    opset_ai_onnx_ml5,
)

from .onnx_types import (
    BFLOAT16,
    FLOAT16,
    FLOAT8E4M3FN,
    FLOAT8E4M3FNUZ,
    FLOAT8E5M2,
    FLOAT8E5M2FNUZ,
    FLOAT,
    DOUBLE,
    INT8,
    INT16,
    INT32,
    INT64,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
    BOOL,
    STRING,
    COMPLEX64,
    COMPLEX128,
)

# isort: on

from . import ir, nn, optimizer, rewriter, version_converter
from ._internal.builder import GraphBuilder, OpBuilder
from ._internal.utils import external_tensor
from ._internal.values import OnnxFunction, TracedOnnxFunction

# Set DEBUG to True to enable additional debug checks
DEBUG: bool = False

try:  # noqa: SIM105
    __version__ = importlib.metadata.version("onnxscript")
except importlib.metadata.PackageNotFoundError:
    # package is not installed
    pass



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import numpy as np

from onnxscript import graph, script
from onnxscript._internal import evaluator
from onnxscript.onnx_opset import opset17 as op
from onnxscript.onnx_types import FLOAT


class EvaluatorTest(unittest.TestCase):
    def test_evaluator(self):
        @script()
        def seq_map(x: FLOAT["N"]):  # noqa: F821
            seq1 = op.SequenceConstruct(x, x + 1, x + 2)

            @graph()
            def square(y: FLOAT["N"]) -> FLOAT["N"]:  # noqa: F821
                return op.Mul(y, y)

            seq2 = op.SequenceMap(seq1, body=square)
            return seq2

        x = np.array([0.0, 1.0], dtype=np.float32)
        expected = [t * t for t in (x, x + 1, x + 2)]

        # Test using (current) default evaluator
        output = seq_map(x)
        np.testing.assert_equal(output, expected)

        # Test using ort-mixed-evaluator
        with evaluator.default_as(evaluator.ort_mixed_evaluator):
            output = seq_map(x)
        np.testing.assert_equal(output, expected)

        # Test using ort-evaluator
        with evaluator.default_as(evaluator.ort_evaluator):
            output = seq_map(x)
        np.testing.assert_equal(output, expected)


class ORTEvaluatorTest(unittest.TestCase):
    def test_it_ignores_unknown_function_kwargs_when_option_set_to_true(self):
        @script()
        def test_function(x, y: float = 1.0):
            return op.Add(x, y)

        x = np.array(0.0, dtype=np.float32)
        expected = np.array(1.0, dtype=np.float32)
        with evaluator.default_as(evaluator.ORTEvaluator(ignore_unknown_function_kwargs=True)):
            output = test_function(x, unknown=42)  # pylint: disable=unexpected-keyword-arg

        np.testing.assert_equal(output, expected)

    def test_it_raise_on_unknown_function_kwargs_by_default(self):
        @script()
        def test_function(x, y: float = 1.0):
            return op.Add(x, y)

        x = np.array(0.0, dtype=np.float32)
        with evaluator.default_as(evaluator.ORTEvaluator()):  # noqa: SIM117
            with self.assertRaises(TypeError):
                _ = test_function(x, unknown=42)  # pylint: disable=unexpected-keyword-arg


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Convenience methods for constructing the IR."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Sequence

from onnx_ir import tape

if TYPE_CHECKING:
    import onnx_ir as ir


# A type representing the domains/versions used in creating nodes in IR.
UsedOpsets = set[tuple[str, Optional[int]]]


class Builder(tape.Tape):
    """An extension of the tape that provides a more convenient API for constructing the IR.

    Example:
        >>> from onnxscript import ir
        >>> from onnxscript.ir import _tape
        >>> op = _tape.Builder()
        >>> input = ir.Value(name="input", type=ir.TensorType(ir.DataType.FLOAT), shape=ir.Shape((1, 2)))
        >>> relu_val = op.Relu(input, _name="relu_node", _domain="", _version=18, _outputs=["relu_out"])

    Note: When passing `_name`, ensure it is unique to avoid duplicate node names.
    """

    def __getattr__(self, op_type: str) -> Any:
        return lambda *args, **kwargs: self._make_node(op_type, args, kwargs)

    def _make_node(self, op_type: str, inputs: Sequence[ir.Value], kwargs: dict[str, Any]):
        domain = kwargs.pop("_domain", "")
        version = kwargs.pop("_version", None)
        outputs = kwargs.pop("_outputs", 1)
        name = kwargs.pop("_name", None)

        if isinstance(outputs, Sequence):
            num_outputs = len(outputs)
        else:
            assert isinstance(outputs, int)
            num_outputs = outputs

        if num_outputs == 1:
            value = super().op(
                op_type,
                inputs=inputs,
                attributes=kwargs,
                domain=domain,
                version=version,
                name=name,
            )
            if isinstance(outputs, Sequence):
                value.name = outputs[0]
            return value
        values = super().op_multi_out(
            op_type,
            inputs=inputs,
            attributes=kwargs,
            domain=domain,
            version=version,
            name=name,
            num_outputs=num_outputs,
        )
        if isinstance(outputs, Sequence):
            for value, name in zip(values, outputs):
                value.name = name
        return values



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import logging

from onnxscript import ir
from onnxscript.rewriter._rewrite_rule import RewriteRule, RewriteRuleSet

logger = logging.getLogger(__name__)


def cast_constant_of_shape(op, shape, scalar, dtype):
    constant = op.ConstantOfShape(shape, value=scalar)
    return op.Cast(constant, to=dtype)


def fused_cast_constant_of_shape(op, shape: ir.Value, scalar: ir.Attr, dtype: ir.Attr, **_):
    # Cast scalar (a TensorProto attribute) to the specified dtype
    scalar_value = scalar.value.numpy().item()
    cast_value = ir.tensor([scalar_value], dtype=ir.DataType(dtype.as_int()))
    return op.ConstantOfShape(shape, value=cast_value)


def cast_constant_of_shape_without_value(op, shape, dtype):
    constant = op.ConstantOfShape(shape)
    return op.Cast(constant, to=dtype)


def fused_cast_constant_of_shape_without_value(op, shape, dtype, **_):
    zero = ir.tensor([0], dtype=ir.DataType(dtype.as_int()))
    return op.ConstantOfShape(shape, value=zero)


cast_constant_of_shape_rule = RewriteRule(cast_constant_of_shape, fused_cast_constant_of_shape)

cast_constant_of_shape_without_value_rule = RewriteRule(
    cast_constant_of_shape_without_value, fused_cast_constant_of_shape_without_value
)

rules = RewriteRuleSet(
    [
        cast_constant_of_shape_rule,
        cast_constant_of_shape_without_value_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

from parameterized import parameterized

import onnxscript.optimizer
from onnxscript.rewriter.models import _rotary_embedding_models, _smollm_1
from onnxscript.rewriter.ort_fusions import rotary_embedding


class TestRotaryEmbedding(unittest.TestCase):
    @parameterized.expand(
        [
            (
                "test_case_1",
                _rotary_embedding_models.test_case_1,
            ),
            (
                "smollm_test_1",
                _smollm_1.smollm_test_1,
            ),
        ]
    )
    def test_rotary_embedding_fusion(self, _: str, test_data_constructor):
        test = test_data_constructor()
        model = test.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        rotary_embedding.fuse_rotary_embedding(model)
        op_types = [n.op_type for n in model.graph]
        self.assertIn("RotaryEmbedding", op_types)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import abc
import contextlib
import pprint
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    runtime_checkable,
)

import numpy as np
import onnx
import onnx.defs
import onnx.reference
import onnx_ir as ir
from typing_extensions import TypeAlias

from onnxscript import onnx_opset, tensor
from onnxscript._internal import autocast, param_manipulation, utils, values

UserModeValue: TypeAlias = Union[Optional[np.ndarray], Sequence["UserModeValue"]]

EagerModeValue: TypeAlias = Union[Optional["tensor.Tensor"], Sequence["EagerModeValue"]]

ExtendedModeValue: TypeAlias = Union[
    Optional["tensor.Tensor"],
    Sequence["ExtendedModeValue"],
    np.ndarray,
    int,
    float,
    bool,
    str,
]

_T = TypeVar("_T")


def _adapt_to_eager_mode(inputs: ExtendedModeValue) -> tuple[EagerModeValue, bool]:
    """Adapts inputs into representation used by onnxscript eager mode.

    This does the following transformations:
    * It adds an onnxscript Tensor wrapper around numpy arrays, which
    allows the use of overloaded operators like + to be controlled by onnxscript.
    * It also provides a promotion of scalars into tensors as a convenience.
    This is needed to complement the similar promotion supported by the
    onnxscript converter (for example, when an attribute is promoted and used
    as an input argument).

    Args:
        inputs: a list/tuple of inputs to an ONNX function

    Returns:
        a pair (wrapped_inputs, flag) where flag indicates whether any numpy array
        was wrapped into a Tensor.
    """
    has_array = False

    def adapt(input: ExtendedModeValue) -> EagerModeValue:
        if isinstance(input, np.ndarray):
            nonlocal has_array
            has_array = True
            return tensor.Tensor(input)
        if isinstance(input, tensor.Tensor):
            return input
        if isinstance(input, (bool, float)):
            return tensor.Tensor(np.array(input))
        if isinstance(input, int):
            return tensor.Tensor(np.array(input, dtype=np.int64))
        if input is None:
            return None
        if isinstance(input, list):
            return [adapt(elt) for elt in input]
        if isinstance(input, tuple):
            return tuple(adapt(elt) for elt in input)
        raise TypeError(f"Unexpected input type {type(input)}.")

    result = adapt(inputs)
    return result, has_array


def _adapt_to_user_mode(output: ExtendedModeValue) -> UserModeValue:
    """Unwraps Tensor wrapper around numpy arrays.

    Args:
        output: output of an ONNX function, which can be either a single
            onnx value or a list/tuple of onnx values.

    Returns:
        unwrapped output
    """
    if isinstance(output, tensor.Tensor):
        return output.value
    if output is None:
        return None
    if isinstance(output, list):
        return [_adapt_to_user_mode(elt) for elt in output]
    if isinstance(output, tuple):
        return tuple(_adapt_to_user_mode(elt) for elt in output)
    if isinstance(output, np.ndarray):
        return output
    raise TypeError(f"Unexpected type {type(output)}.")


def _unwrap_tensors_in_kwargs(kwargs: Mapping[str, Any]) -> dict[str, Any]:
    """Unwrap tensors in a mapping to numpy arrays."""
    new_kwargs = {}
    for k, v in kwargs.items():
        new_kwargs[k] = v
        if isinstance(v, tensor.Tensor):
            new_kwargs[k] = v.value

    return new_kwargs


@runtime_checkable
class Evaluator(Protocol):
    """Protocol for evaluating ONNX ops.

    NOTE: The ``eval`` method was deprecated and removed. Implement ``eval_op``
    and ``eval_function`` instead.
    """

    def eval_op(
        self,
        op: values.Op,
        args: Sequence[ExtendedModeValue],
        kwargs: Mapping[str, ExtendedModeValue],
    ):
        """Evaluates an Op.

        Args:
            op: The Op to evaluate.
            args: The positional arguments to the op.
            kwargs: The keyword arguments to the op.
        """

    def eval_function(
        self,
        function: values.OnnxFunction,
        args: Sequence[ExtendedModeValue],
        kwargs: Mapping[str, ExtendedModeValue],
    ):
        """Evaluates an OnnxFunction.

        Args:
            function: The OnnxFunction to evaluate.
            args: The positional arguments to the function.
            kwargs: The keyword arguments to the function.
        """


class BaseEvaluator(Evaluator, abc.ABC):
    """Base class for evaluation of ONNX ops.

    The execution of onnxscript functions in eager-mode is dispatched to an Evaluator
    instance (or, more precisely, to the eval method of the Evaluator instance).
    The evaluator is expected to transform the input/output/attribute representation
    supported by onnxscript to those expected by a particular backend.
    """

    def __init__(self, ignore_unknown_function_kwargs: bool = False):
        """Initializes a BaseEvaluator.

        Args:
            ignore_unknown_function_kwargs: Whether to ignore unknown keyword arguments
                when evaluating an OnnxFunction. This is useful when using the
                evaluator to validate operators programmatically, where
                additional keyword arguments that is not part of the signature
                may be provided to the function.
        """
        self._ignore_unknown_function_kwargs = ignore_unknown_function_kwargs

    def _adapt_inputs(
        self, op_signature: ir.schemas.OpSignature, inputs: Sequence[ExtendedModeValue]
    ):
        """Transform inputs to the expected format for the evaluator.

        Enables some syntactic sugar, such as the use of Python scalars,
        in a manner consistent with the translator. See autocast.py for details.
        """
        return autocast.dynamic_cast_inputs(op_signature, inputs)

    def _adapt_attributes(
        self, op_signature, attributes: Mapping[str, ExtendedModeValue]
    ) -> tuple[dict[str, ExtendedModeValue], dict[str, ExtendedModeValue]]:
        """Transform attributes to the expected format for the evaluator.

        Returns:
            A closure that can be used to evaluate graph-valued attributes.
        """
        use_graph_attribute = self.use_graph_attribute(op_signature)
        closure: dict[Any, Any] = {}
        adapted_attributes = {}
        for k, v in attributes.items():
            if isinstance(v, values.OnnxClosure):
                if use_graph_attribute:
                    adapted_attributes[k] = v.function_ir.to_graph_proto()
                    for pyvar, onnxvar in v.function_ir.outer_scope_variables:
                        closure[onnxvar.value.name] = v.frame.f_locals[pyvar]
                else:
                    adapted_attributes[k] = v.function
            elif callable(v):
                raise TypeError(
                    f"Error: function-valued attribute {v.__name__} has no graph_proto"
                    "attribute. Did you forget to decorate it with @graph?"
                )
            else:
                adapted_attributes[k] = v
        return adapted_attributes, closure

    def _adapt_outputs(self, outputs: Sequence[EagerModeValue]):
        """Adapt evaluator's output to convention used in onnxscript.

        Onnxscript uses a tuple/sequence only when number of outputs > 1.
        """
        return outputs[0] if len(outputs) == 1 else outputs

    def use_graph_attribute(self, op_signature: ir.schemas.OpSignature) -> bool:
        del op_signature  # unused
        return True

    @abc.abstractmethod
    def _eval(
        self,
        schema: onnx.defs.OpSchema,
        inputs: Sequence[ExtendedModeValue],
        attributes: Mapping[str, ExtendedModeValue],
        closure: Mapping[str, ExtendedModeValue],
    ) -> EagerModeValue:
        """Evaluates an ONNX op given its schema and inputs/attributes.

        Args:
            schema: The schema of the op to evaluate.
            inputs: The ONNX inputs to the op.
            attributes: The ONNX attributes to the op.
            closure: The closure to use when evaluating graph-valued attributes.
        """

    def eval_op(
        self,
        op: values.Op,
        args: Sequence[ExtendedModeValue],
        kwargs: Mapping[str, ExtendedModeValue],
    ):
        op_signature = op.op_signature
        assert op_signature is not None, f"Op {op.name} has no signature."
        attributes = _unwrap_tensors_in_kwargs(kwargs)
        attributes, closure = self._adapt_attributes(op_signature, attributes)
        inputs = self._adapt_inputs(op_signature, args)
        outputs = self._eval(op.op_schema, inputs, attributes, closure)
        return self._adapt_outputs(outputs)

    def eval_function(
        self,
        function: values.OnnxFunction,
        args: Sequence[ExtendedModeValue],
        kwargs: Mapping[str, ExtendedModeValue],
    ):
        """Evaluates a function in eager mode.

        Override this function to change the evaluator's behavior for functions.

        Args:
            function: The OnnxFunction to evaluate.
            args: The positional arguments to the function.
            kwargs: The keyword arguments to the function.
        """
        op_signature = function.op_signature
        if op_signature is None:
            raise RuntimeError(f"Function {function.name} has no signature.")
        # Split happens in the evaluator instead of the OnnxFunction __call__ method
        # so that evaluators can control behaviors like whether to fill in default values for attributes.
        tagged_args, tagged_kwargs = param_manipulation.tag_arguments_with_signature(
            op_signature,
            args,
            kwargs,
            fill_defaults=False,
            allow_extra_kwargs=self._ignore_unknown_function_kwargs,
        )

        adapted_args: list[ExtendedModeValue] = []
        adapted_kwargs: dict[str, ExtendedModeValue] = {}
        has_array = False
        for arg, param in tagged_args:
            if isinstance(param, ir.schemas.Parameter):
                adapted_arg, has_array_ = _adapt_to_eager_mode(arg)
                has_array = has_array or has_array_
                adapted_args.append(adapted_arg)
            else:
                adapted_args.append(arg)

        for key, (arg, param) in tagged_kwargs.items():
            if isinstance(param, ir.schemas.Parameter):
                adapted_arg, has_array_ = _adapt_to_eager_mode(arg)
                has_array = has_array or has_array_
                adapted_kwargs[key] = adapted_arg
            else:
                adapted_kwargs[key] = arg

        result = function.function(*adapted_args, **adapted_kwargs)

        # We use a heuristic to decide whether to return output values as
        # numpy arrays or tensor.Tensors. If the function has at least one
        # numpy array as input, we return numpy arrays. Otherwise, we return
        # tensor.Tensors. We could use a user-specified flag to control this
        # or explicitly track whether this is a top-level function-call or
        # a nested function-call.

        return _adapt_to_user_mode(result) if has_array else result


# Utilities for evaluation using ORT:


class EagerModeError(RuntimeError):
    pass


def _rename_io(prefix, i, arg):
    if arg is None:
        return ""
    return f"{prefix}{i}"


def compute_num_outputs(
    schema: onnx.defs.OpSchema, args: Sequence[Any], kwargs: Mapping[str, Any]
) -> int:
    """Returns the number of outputs expected."""

    # TODO: Use ONNX type inference to replace the special-case handling below.
    if schema.domain == "":
        if schema.name == "BatchNormalization":
            if not kwargs.get("training_mode", 0):
                return 1
        if schema.name == "LSTM":
            return 3
        if schema.name == "Split":
            if len(args) == 1 and "num_outputs" not in kwargs:
                raise EagerModeError(
                    "Operator Split: the number of expected outputs defines the split. "
                    "This information is unknown here."
                )
            if len(args) == 2:  # has argument(split)
                return len(args[1])
            else:  # no argument(split), check attribute(num_outputs)
                return kwargs["num_outputs"]
        if schema.name == "Scan":
            scan_body = kwargs["body"]
            return len(scan_body.output)
        if schema.name == "Loop":
            loop_body = kwargs["body"]
            return len(loop_body.output) - 1
    return len(schema.outputs)


def _onnxscript_to_numpy_value(v):
    """Converts an onnxscript encoding of an ONNX value into the numpy encoding used by runtimes."""
    if isinstance(v, tensor.Tensor):
        return v.value
    if isinstance(v, list):
        return [_onnxscript_to_numpy_value(x) for x in v]
    if isinstance(v, tuple):
        if len(v) > 0 and type(v[0]) is int:  # pylint: disable=unidiomatic-typecheck
            return np.array(v, dtype=np.int64)
        return np.array(v)
    if v is None:
        # Treated as a static-optional value.
        # Dynamic optional None not yet supported.
        return v
    if isinstance(v, np.ndarray):
        return v
    raise TypeError(
        f"Unexpected onnxscript value type '{type(v)}'."
        "Valid value types are 'Tensor | list[Tensor] | None | np.ndarray | list[np.ndarray]'"
    )


def _numpy_to_onnxscript_value(
    v: np.ndarray | np.generic | list[np.ndarray] | list[np.generic],
):
    """Converts an ORT encoding of an ONNX value into the encoding used by onnxscript."""
    if isinstance(v, np.ndarray):
        # ORT may reuse buffers when the output numpy array is provided back as input.
        # We need to make a copy to ensure that the tensor is not modified in-place.
        return tensor.Tensor(v.copy())
    if issubclass(type(v), np.generic):
        # Numpy scalar types that are not ndarray
        # https://numpy.org/doc/stable/reference/arrays.scalars.html
        return tensor.Tensor(np.array(v))
    if isinstance(v, list):
        return [_numpy_to_onnxscript_value(x) for x in v]
    if v is None:
        raise TypeError("Dynamic optional values not yet supported.")
    raise TypeError(
        f"Unexpected runtime value type '{type(v)}'."
        "Valid types are: 'np.ndarray | np.generic | list[np.ndarray] | list[np.generic]'"
    )


def _prepare_model_and_inputs_for_eager(
    schema: onnx.defs.OpSchema,
    args: Sequence[Any],
    kwargs: Mapping[str, Any],
    implicit_args: Optional[Mapping[str, Any]],
):
    implicit_args = implicit_args or {}
    # Convert input values to ORT representation-type:
    args = [_onnxscript_to_numpy_value(x) for x in args]
    implicit_args = {k: _onnxscript_to_numpy_value(v) for k, v in implicit_args.items()}

    # Utility to convert kwarg to ONNX AttributeProto:
    # TODO(justinchuby): Clean up this function to use onnx-ir
    def make_attr(key: str, value: Any) -> onnx.AttributeProto:
        def make_tensor_name() -> str:
            return f"attr_{key}"

        return ir.to_proto(
            autocast.pyvalue_to_onnx_attribute(
                key, value, make_tensor_name, int(schema.attributes[key].type)
            )
        )

    # Construct ONNX model with a single op call:
    inputs = [_rename_io("input", i, arg) for i, arg in enumerate(args)]

    num_outputs = compute_num_outputs(schema, args, kwargs)
    outputs = [f"output{i}" for i in range(num_outputs)]

    node = onnx.helper.make_node(schema.name, inputs, outputs, domain=schema.domain)  # noqa: TID251
    node.attribute.extend(
        make_attr(key, value) for key, value in kwargs.items() if value is not None
    )
    input_value_infos = utils.values_to_value_infos(zip(inputs, args))
    implicit_value_infos = utils.values_to_value_infos(implicit_args.items())
    output_value_infos = [
        onnx.helper.make_value_info(name, onnx.TypeProto())  # noqa: TID251
        for name in outputs
    ]

    graph = onnx.helper.make_graph(  # noqa: TID251
        [node], "node_graph", input_value_infos + implicit_value_infos, output_value_infos
    )
    opset_id = onnx.helper.make_opsetid(schema.domain, schema.since_version)  # noqa: TID251
    model = onnx.helper.make_model(  # noqa: TID251
        graph,
        opset_imports=[opset_id],
        ir_version=values.select_ir_version(schema.since_version, domain=schema.domain),
    )
    model = onnx.shape_inference.infer_shapes(model)

    session_run_input = {name: arg for name, arg in zip(inputs, args) if name != ""}
    session_run_input.update(implicit_args)

    return model, session_run_input, inputs


def _call_ort(
    schema: onnx.defs.OpSchema,
    args: Sequence[Any],
    kwargs: Mapping[str, Any],
    implicit_args: Optional[Mapping[str, Any]],
):
    # Delay import onnxruntime so that onnxscript can be used without
    # installing onnxruntime.
    import onnxruntime as ort  # pylint: disable=import-outside-toplevel
    from onnxruntime.capi.onnxruntime_pybind11_state import (  # pylint: disable=import-outside-toplevel
        Fail,
        InvalidArgument,
        InvalidGraph,
    )

    model, session_run_input, inputs = _prepare_model_and_inputs_for_eager(
        schema, args, kwargs, implicit_args
    )

    try:
        session = ort.InferenceSession(
            model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
    except (Fail, InvalidGraph, InvalidArgument) as e:
        raise EagerModeError(
            f"Unable to create onnxruntime InferenceSession "
            f"for executing {schema.domain}.{schema.name} op "
            f"with onnx model\n{onnx.printer.to_text(model)}"
        ) from e

    try:
        result = session.run(None, session_run_input)
    except (RuntimeError, Fail) as e:
        raise EagerModeError(
            f"Unable to execute model operator {schema.name!r} due to {e!r}"
            f"\ninput types:\n"
            f"{pprint.pformat({k: type(v) for k, v in zip(inputs, args)})}"
            f"\nmodified input types:\n"
            f"{pprint.pformat({k: type(v) for k, v in session_run_input.items()})}"
            f"\ninputs:\n{pprint.pformat(session_run_input)}\n{model}"
        ) from e

    # Map ORT output values to the onnxscript representation-type.
    return [_numpy_to_onnxscript_value(x) for x in result]


def _op_identifier(
    op_schema_or_signature: onnx.defs.OpSchema | ir.schemas.OpSignature,
) -> tuple[str, str, int]:
    return (
        op_schema_or_signature.name,
        op_schema_or_signature.domain,
        op_schema_or_signature.since_version,
    )


class ORTEvaluator(BaseEvaluator):
    """Evaluates ONNX ops using ONNX Runtime."""

    def _eval(self, schema, inputs, attributes, closure):
        return _call_ort(schema, inputs, attributes, closure)


class OnnxReferenceRuntimeEvaluator(BaseEvaluator):
    """Evaluates ONNX ops using ONNX Runtime."""

    def _eval(self, schema, inputs, attributes, closure):
        model, session_run_input, adapted_inputs = _prepare_model_and_inputs_for_eager(
            schema, inputs, attributes, closure
        )
        session = onnx.reference.ReferenceEvaluator(model)
        try:
            result = session.run(None, session_run_input)
        except RuntimeError as e:
            raise EagerModeError(
                f"Unable to execute model operator {schema.name!r} due to {e!r}"
                f"\ninput types:\n"
                f"{pprint.pformat({k: type(v) for k, v in zip(adapted_inputs, inputs)})}"
                f"\nmodified input types:\n"
                f"{pprint.pformat({k: type(v) for k, v in session_run_input.items()})}"
                f"\ninputs:\n{pprint.pformat(session_run_input)}\n{model}"
            ) from e

        return [_numpy_to_onnxscript_value(x) for x in result]


ort_evaluator = ORTEvaluator()


class ORTMixedEvaluator(ORTEvaluator):
    """Evaluates ONNX ops using ONNX Runtime, unless an overriding python implementation is registered.

    This is useful for higher-order ops such as Scan and SequenceMap, allowing for
    python-based debugging.
    """

    def __init__(self) -> None:
        super().__init__()
        self._python_ops: dict[tuple[str, str, int], Any] = {}

    def use_graph_attribute(self, op_signature: ir.schemas.OpSignature) -> bool:
        return _op_identifier(op_signature) not in self._python_ops

    def _eval(self, schema, inputs, attributes, closure):
        identifier = _op_identifier(schema)
        if identifier in self._python_ops:
            return self._python_ops[identifier](inputs, attributes)
        else:
            return super()._eval(schema, inputs, attributes, closure)

    def register(self, opset: values.Opset) -> Callable[[_T], _T]:
        assert opset is not None

        def decorator(function: _T) -> _T:
            op_signature = opset[function.__name__].op_signature
            self._python_ops[_op_identifier(op_signature)] = function
            return function

        return decorator


ort_mixed_evaluator = ORTMixedEvaluator()


@ort_mixed_evaluator.register(opset=onnx_opset.opset18)
def SequenceMap(inputs: Sequence[Any], attributes: Mapping[str, Any]):
    """Evaluates a SequenceMap op."""
    fun = attributes["body"]

    def get_input_of(input_index, iter_num):
        input = inputs[input_index]
        if isinstance(input, list):
            return input[iter_num]
        return input

    def get_input(iter_num):
        return [get_input_of(input_index, iter_num) for input_index in range(len(inputs))]

    return [fun(*(get_input(i))) for i in range(len(inputs[0]))]


# Used to control the default evaluator instance. A simple approach for now.

_default_evaluator: Evaluator = ort_evaluator


def default() -> Evaluator:
    """Returns the default Evaluator default."""
    return _default_evaluator


def set_default(new_default: Evaluator) -> None:
    """Sets the current Evaluator default."""
    global _default_evaluator  # pylint: disable=global-statement
    _default_evaluator = new_default


@contextlib.contextmanager
def default_as(temp_default: Evaluator):
    """Context manager that temporarily switches the default evaluator."""
    old_default = _default_evaluator
    set_default(temp_default)
    try:
        yield
    finally:
        set_default(old_default)
